from typing import Optional

import pytest
from pydantic import ConfigDict

from ninja import Body, Cookie, Header, Router, Schema
from ninja.testing import TestClient


class OptionalEmptySchema(Schema):
    model_config = ConfigDict(extra="forbid")
    name: Optional[str] = None


class ExtraForbidSchema(Schema):
    model_config = ConfigDict(extra="forbid")
    name: str
    metadata: Optional[OptionalEmptySchema] = None


router = Router()


@router.get("/headers1")
def headers1(request, user_agent: str = Header(...)):
    return user_agent


@router.get("/headers2")
def headers2(request, ua: str = Header(..., alias="User-Agent")):
    return ua


@router.get("/headers3")
def headers3(request, content_length: int = Header(...)):
    return content_length


@router.get("/headers4")
def headers4(request, c_len: int = Header(..., alias="Content-length")):
    return c_len


@router.get("/headers5")
def headers5(request, missing: int = Header(...)):
    return missing


@router.get("/cookies1")
def cookies1(request, weapon: str = Cookie(...)):
    return weapon


@router.get("/cookies2")
def cookies2(request, wpn: str = Cookie(..., alias="weapon")):
    return wpn


@router.post("/test-schema")
def schema(request, payload: ExtraForbidSchema = Body(...)):
    return "ok"


client = TestClient(router)


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/headers1", 200, "Ninja"),
        ("/headers2", 200, "Ninja"),
        ("/headers3", 200, 10),
        ("/headers4", 200, 10),
        (
            "/headers5",
            422,
            {
                "detail": [
                    {
                        "type": "missing",
                        "loc": ["header", "missing"],
                        "msg": "Field required",
                    }
                ]
            },
        ),
        ("/cookies1", 200, "shuriken"),
        ("/cookies2", 200, "shuriken"),
    ],
)
def test_headers(path, expected_status, expected_response):
    response = client.get(
        path,
        headers={"User-Agent": "Ninja", "Content-Length": "10"},
        COOKIES={"weapon": "shuriken"},
    )
    assert response.status_code == expected_status, response.content
    print(response.json())
    assert response.json() == expected_response


@pytest.mark.parametrize(
    "path,json,expected_status,expected_response",
    [
        (
            "/test-schema",
            {"name": "test", "extra_name": "test2"},
            422,
            {
                "detail": [
                    {
                        "type": "extra_forbidden",
                        "loc": ["body", "payload", "extra_name"],
                        "msg": "Extra inputs are not permitted",
                    }
                ]
            },
        ),
        (
            "/test-schema",
            {"name": "test", "metadata": {"extra_name": "xxx"}},
            422,
            {
                "detail": [
                    {
                        "loc": ["body", "payload", "metadata", "extra_name"],
                        "msg": "Extra inputs are not permitted",
                        "type": "extra_forbidden",
                    }
                ]
            },
        ),
        (
            "/test-schema",
            {"name": "test", "metadata": "test2"},
            422,
            {
                "detail": [
                    {
                        "type": "model_attributes_type",
                        "loc": ["body", "payload", "metadata"],
                        "msg": "Input should be a valid dictionary or object to extract fields from",
                    }
                ]
            },
        ),
    ],
)
def test_pydantic_config(path, json, expected_status, expected_response):
    # test extra forbid
    response = client.post(path, json=json)
    assert response.json() == expected_response
    assert response.status_code == expected_status



from abc import ABC, abstractmethod
from typing import Any, Optional

from django.http import HttpRequest

from ninja.errors import HttpError
from ninja.security.base import AuthBase
from ninja.utils import check_csrf

__all__ = ["APIKeyBase", "APIKeyQuery", "APIKeyCookie", "APIKeyHeader"]


class APIKeyBase(AuthBase, ABC):
    openapi_type: str = "apiKey"
    param_name: str = "key"

    def __init__(self) -> None:
        self.openapi_name = self.param_name  # this sets the name of the security schema
        super().__init__()

    def __call__(self, request: HttpRequest) -> Optional[Any]:
        key = self._get_key(request)
        return self.authenticate(request, key)

    @abstractmethod
    def _get_key(self, request: HttpRequest) -> Optional[str]:
        pass  # pragma: no cover

    @abstractmethod
    def authenticate(self, request: HttpRequest, key: Optional[str]) -> Optional[Any]:
        pass  # pragma: no cover


class APIKeyQuery(APIKeyBase, ABC):
    openapi_in: str = "query"

    def _get_key(self, request: HttpRequest) -> Optional[str]:
        return request.GET.get(self.param_name)


class APIKeyCookie(APIKeyBase, ABC):
    openapi_in: str = "cookie"

    def __init__(self, csrf: bool = True) -> None:
        self.csrf = csrf
        super().__init__()

    def _get_key(self, request: HttpRequest) -> Optional[str]:
        # Skip CSRF check if the operation is marked as csrf_exempt
        if self.csrf and not getattr(request, "_ninja_csrf_exempt", False):
            error_response = check_csrf(request)
            if error_response:
                raise HttpError(403, "CSRF check Failed")
        return request.COOKIES.get(self.param_name)


class APIKeyHeader(APIKeyBase, ABC):
    openapi_in: str = "header"

    def _get_key(self, request: HttpRequest) -> Optional[str]:
        headers = request.headers
        return headers.get(self.param_name)



from typing import List, Union

import pytest
from pydantic import ValidationError

from ninja import NinjaAPI, Schema
from ninja.errors import ConfigError
from ninja.responses import codes_2xx, codes_3xx
from ninja.testing import TestClient

api = NinjaAPI()


@api.get("/check_int", response={200: int})
def check_int(request):
    return 200, "1"


@api.get("/check_int2", response={200: int})
def check_int2(request):
    return 200, "str"


@api.get("/check_single_with_status", response=int)
def check_single_with_status(request, code: int):
    return code, 1


@api.get("/check_response_schema", response={400: int})
def check_response_schema(request):
    return 200, 1


@api.get("/check_no_content", response={204: None})
def check_no_content(request, return_code: bool):
    if return_code:
        return 204, None
    return  # None


@api.get(
    "/check_multiple_codes",
    response={codes_2xx: int, codes_3xx: str, ...: float},
)
def check_multiple_codes(request, code: int):
    return code, "1"


class User:
    def __init__(self, id, name, password):
        self.id = id
        self.name = name
        self.password = password


class UserModel(Schema):
    id: int
    name: str
    # skipping password output to responses


class ErrorModel(Schema):
    detail: str


@api.get("/check_model", response={200: UserModel, 202: UserModel})
def check_model(request):
    return 202, User(1, "John", "Password")


@api.get("/check_list_model", response={200: List[UserModel]})
def check_list_model(request):
    return 200, [User(1, "John", "Password")]


@api.get("/check_union", response={200: Union[int, UserModel], 400: ErrorModel})
def check_union(request, q: int):
    if q == 0:
        return 200, 1
    if q == 1:
        return 200, User(1, "John", "Password")
    if q == 2:
        return 400, {"detail": "error"}
    return "invalid"


client = TestClient(api)


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/check_int", 200, 1),
        ("/check_single_with_status?code=200", 200, 1),
        ("/check_model", 202, {"id": 1, "name": "John"}),  # ! the password is skipped
        ("/check_list_model", 200, [{"id": 1, "name": "John"}]),
        ("/check_union?q=0", 200, 1),
        ("/check_union?q=1", 200, {"id": 1, "name": "John"}),
        ("/check_union?q=2", 400, {"detail": "error"}),
        ("/check_multiple_codes?code=200", 200, 1),
        ("/check_multiple_codes?code=201", 201, 1),
        ("/check_multiple_codes?code=202", 202, 1),
        ("/check_multiple_codes?code=206", 206, 1),
        ("/check_multiple_codes?code=300", 300, "1"),
        ("/check_multiple_codes?code=308", 308, "1"),
        ("/check_multiple_codes?code=400", 400, 1.0),
        ("/check_multiple_codes?code=500", 500, 1.0),
    ],
)
def test_responses(path, expected_status, expected_response):
    response = client.get(path)
    assert response.status_code == expected_status, response.content
    assert response.json() == expected_response


def test_schema():
    checks = [
        ("/api/check_int", {200}),
        ("/api/check_int2", {200}),
        ("/api/check_single_with_status", {200}),
        ("/api/check_response_schema", {400}),
        ("/api/check_model", {200, 202}),
        ("/api/check_list_model", {200}),
        ("/api/check_union", {200, 400}),
    ]
    schema = api.get_openapi_schema()

    # checking codes
    for path, codes in checks:
        responses = schema["paths"][path]["get"]["responses"]
        responses_codes = set(responses.keys())
        assert codes == responses_codes, f"{codes} != {responses_codes}"

    # checking model
    check_model_responses = schema["paths"]["/api/check_model"]["get"]["responses"]

    assert check_model_responses == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/UserModel"}
                }
            },
            "description": "OK",
        },
        202: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/UserModel"}
                }
            },
            "description": "Accepted",
        },
    }


def test_no_content():
    response = client.get("/check_no_content?return_code=1")
    assert response.status_code == 204
    assert response.content == b""

    response = client.get("/check_no_content?return_code=0")
    assert response.status_code == 204
    assert response.content == b""

    schema = api.get_openapi_schema()
    details = schema["paths"]["/api/check_no_content"]["get"]["responses"]
    assert details == {204: {"description": "No Content"}}


def test_validates():
    with pytest.raises(ValidationError):
        client.get("/check_int2")

    with pytest.raises(ValidationError):
        client.get("/check_union?q=3")

    with pytest.raises(ConfigError):
        client.get("/check_response_schema")

    with pytest.raises(ConfigError):
        client.get("/check_single_with_status?code=300")



# This is no longer the case in pydantic 2
# https://github.com/pydantic/pydantic/issues/5991
# from datetime import date
# from typing import Union

# from ninja import Router
# from ninja.testing import TestClient

# router = Router()


# @router.get("/test")
# def view(request, value: Union[date, str]):
#     return [value, type(value).__name__]


# client = TestClient(router)


# def test_union():
#     assert client.get("/test?value=today").json() == ["today", "str"]
#     assert client.get("/test?value=2020-01-15").json() == ["2020-01-15", "date"]



import inspect
import warnings
from collections import defaultdict, namedtuple
from sys import version_info
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple

import pydantic
from django.http import HttpResponse
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from typing_extensions import Annotated, get_args, get_origin

from ninja import UploadedFile
from ninja.compatibility.util import UNION_TYPES
from ninja.errors import ConfigError
from ninja.params.models import (
    Body,
    File,
    Form,
    Param,
    Path,
    Query,
    TModel,
    TModels,
    _MultiPartBody,
)
from ninja.signature.utils import get_path_param_names, get_typed_signature
from ninja.utils import is_optional_type

__all__ = [
    "ViewSignature",
    "is_pydantic_model",
    "is_collection_type",
    "detect_collection_fields",
]

FuncParam = namedtuple(
    "FuncParam", ["name", "alias", "source", "annotation", "is_collection"]
)


class ViewSignature:
    FLATTEN_PATH_SEP = (
        "\x1e"  # ASCII Record Separator.  IE: not generally used in query names
    )
    response_arg: Optional[str] = None

    def __init__(self, path: str, view_func: Callable[..., Any]) -> None:
        self.view_func = view_func
        self.signature = get_typed_signature(self.view_func)
        self.path = path
        self.path_params_names = get_path_param_names(path)
        self.docstring = inspect.cleandoc(view_func.__doc__ or "")
        self.has_kwargs = False

        self.params = []
        for name, arg in self.signature.parameters.items():
            if name == "request":
                # TODO: maybe better assert that 1st param is request or check by type?
                # maybe even have attribute like `has_request`
                # so that users can ignore passing request if not needed
                continue

            if arg.kind == arg.VAR_KEYWORD:
                # Skipping **kwargs
                self.has_kwargs = True
                continue

            if arg.kind == arg.VAR_POSITIONAL:
                # Skipping *args
                continue

            if arg.annotation is HttpResponse:
                self.response_arg = name
                continue

            if (
                arg.annotation is inspect.Parameter.empty
                and isinstance(arg.default, type)
                and issubclass(arg.default, pydantic.BaseModel)
            ):
                raise ConfigError(
                    f"Looks like you are using `{name}={arg.default.__name__}` instead of `{name}: {arg.default.__name__}` (annotation)"
                )

            func_param = self._get_param_type(name, arg)
            self.params.append(func_param)

        if hasattr(view_func, "_ninja_contribute_args"):
            # _ninja_contribute_args is a special attribute
            # which allows developers to create custom function params
            # inside decorators or other functions
            for p_name, p_type, p_source in view_func._ninja_contribute_args:
                self.params.append(
                    FuncParam(p_name, p_source.alias or p_name, p_source, p_type, False)
                )

        self.models: TModels = self._create_models()

        self._validate_view_path_params()

    def _validate_view_path_params(self) -> None:
        """verify all path params are present in the path model fields"""
        if self.path_params_names:
            path_model = next(
                (m for m in self.models if m.__ninja_param_source__ == "path"), None
            )
            missing = tuple(
                sorted(
                    name
                    for name in self.path_params_names
                    if not (path_model and name in path_model.__ninja_flatten_map__)
                )
            )
            if missing:
                warnings.warn_explicit(
                    UserWarning(
                        f"Field(s) {missing} are in the view path, but were not found in the view signature."
                    ),
                    category=None,
                    filename=inspect.getfile(self.view_func),
                    lineno=inspect.getsourcelines(self.view_func)[1],
                    source=None,
                )

    def _create_models(self) -> TModels:
        params_by_source_cls: Dict[Any, List[FuncParam]] = defaultdict(list)
        for param in self.params:
            param_source_cls = type(param.source)
            params_by_source_cls[param_source_cls].append(param)

        is_multipart_response_with_body = Body in params_by_source_cls and (
            File in params_by_source_cls or Form in params_by_source_cls
        )
        if is_multipart_response_with_body:
            params_by_source_cls[_MultiPartBody] = params_by_source_cls.pop(Body)

        result = []
        for param_cls, args in params_by_source_cls.items():
            cls_name: str = param_cls.__name__ + "Params"
            attrs = {i.name: i.source for i in args}
            attrs["__ninja_param_source__"] = param_cls._param_source()
            attrs["__ninja_flatten_map_reverse__"] = {}

            if attrs["__ninja_param_source__"] == "file":
                pass

            elif attrs["__ninja_param_source__"] in {
                "form",
                "query",
                "header",
                "cookie",
                "path",
            }:
                flatten_map = self._args_flatten_map(args)
                attrs["__ninja_flatten_map__"] = flatten_map
                attrs["__ninja_flatten_map_reverse__"] = {
                    v: (k,) for k, v in flatten_map.items()
                }

            else:
                assert attrs["__ninja_param_source__"] == "body"
                if is_multipart_response_with_body:
                    attrs["__ninja_body_params__"] = {
                        i.alias: i.annotation for i in args
                    }
                else:
                    # ::TODO:: this is still sus.  build some test cases
                    attrs["__read_from_single_attr__"] = (
                        args[0].name if len(args) == 1 else None
                    )

            # adding annotations
            attrs["__annotations__"] = {i.name: i.annotation for i in args}

            # collection fields:
            attrs["__ninja_collection_fields__"] = detect_collection_fields(
                args, attrs.get("__ninja_flatten_map__", {})
            )

            base_cls = param_cls._model
            model_cls = type(cls_name, (base_cls,), attrs)
            # TODO: https://pydantic-docs.helpmanual.io/usage/models/#dynamic-model-creation - check if anything special in create_model method that I did not use
            result.append(model_cls)
        return result

    def _args_flatten_map(self, args: List[FuncParam]) -> Dict[str, Tuple[str, ...]]:
        flatten_map = {}
        arg_names: Any = {}
        for arg in args:
            if is_pydantic_model(arg.annotation):
                for name, path in self._model_flatten_map(arg.annotation, arg.alias):
                    if name in flatten_map:
                        raise ConfigError(
                            f"Duplicated name: '{name}' in params: '{arg_names[name]}' & '{arg.name}'"
                        )
                    flatten_map[name] = tuple(path.split(self.FLATTEN_PATH_SEP))
                    arg_names[name] = arg.name
            else:
                name = arg.alias
                if name in flatten_map:
                    raise ConfigError(
                        f"Duplicated name: '{name}' also in '{arg_names[name]}'"
                    )
                flatten_map[name] = (name,)
                arg_names[name] = name

        return flatten_map

    def _model_flatten_map(self, model: TModel, prefix: str) -> Generator:
        field: FieldInfo
        for attr, field in model.model_fields.items():
            field_name = field.alias or attr
            name = f"{prefix}{self.FLATTEN_PATH_SEP}{field_name}"
            if is_pydantic_model(field.annotation):
                yield from self._model_flatten_map(field.annotation, name)  # type: ignore
            else:
                yield field_name, name

    def _get_param_type(self, name: str, arg: inspect.Parameter) -> FuncParam:
        # _EMPTY = self.signature.empty
        annotation = arg.annotation
        default = arg.default

        if get_origin(annotation) is Annotated:
            args = get_args(annotation)
            if isinstance(args[-1], Param):
                prev_default = default
                if len(args) == 2:
                    annotation, default = args
                else:
                    # TODO: Remove version check once support for <=3.8 is dropped.
                    # Annotated[] is only available at runtime in 3.9+ per
                    # https://docs.python.org/3/library/typing.html#typing.Annotated
                    if version_info >= (3, 9):
                        # NOTE: Annotated[args[:-1]] seems to have the same runtime
                        # behavior as Annotated[*args[:-1]], but the latter is
                        # invalid in Python < 3.11 because star expressions
                        # were not allowed in index expressions.
                        annotation, default = Annotated[args[:-1]], args[-1]
                    else:  # pragma: no cover -- requires specific Python versions
                        raise NotImplementedError(
                            "This definition requires Python version 3.9+"
                        )
                if prev_default != self.signature.empty:
                    default.default = prev_default

        if annotation == self.signature.empty:
            if default == self.signature.empty:
                annotation = str
            else:
                if isinstance(default, Param):
                    annotation = type(default.default)
                else:
                    annotation = type(default)

            if annotation == PydanticUndefined.__class__:
                # TODO: ^ check why is that so
                annotation = str

        if annotation == type(None) or annotation == type(Ellipsis):  # noqa
            annotation = str

        is_collection = is_collection_type(annotation)

        if annotation == UploadedFile or (
            is_collection and annotation.__args__[0] == UploadedFile
        ):
            # People often forgot to mark UploadedFile as a File, so we better assign it automatically
            if default == self.signature.empty or default is None:
                default = default == self.signature.empty and ... or default
                return FuncParam(name, name, File(default), annotation, is_collection)

        # 1) if type of the param is defined as one of the Param's subclasses - we just use that definition
        if isinstance(default, Param):
            param_source = default

        # 2) if param name is a part of the path parameter
        elif name in self.path_params_names:
            assert (
                default == self.signature.empty
            ), f"'{name}' is a path param, default not allowed"
            param_source = Path(...)

        # 3) if param is a collection, or annotation is part of pydantic model:
        elif is_collection or is_pydantic_model(annotation):
            if default == self.signature.empty:
                param_source = Body(...)
            else:
                param_source = Body(default)

        # 4) the last case is query param
        else:
            if default == self.signature.empty:
                param_source = Query(...)
            else:
                param_source = Query(default)

        # If default is None but annotation is not Optional,
        # wrap it in Optional to allow None values in Pydantic v2
        if default is None and not is_optional_type(annotation):
            annotation = Optional[annotation]

        return FuncParam(
            name, param_source.alias or name, param_source, annotation, is_collection
        )


def is_pydantic_model(cls: Any) -> bool:
    try:
        origin = get_origin(cls)

        # Handle Annotated types - extract the actual type
        if origin is Annotated:
            args = get_args(cls)
            return is_pydantic_model(args[0])

        # Handle Union types
        if origin in UNION_TYPES:
            return any(issubclass(arg, pydantic.BaseModel) for arg in get_args(cls))
        return issubclass(cls, pydantic.BaseModel)
    except TypeError:  # pragma: no cover
        return False


def is_collection_type(annotation: Any) -> bool:
    origin = get_origin(annotation)

    if origin in UNION_TYPES:
        for arg in get_args(annotation):
            if is_collection_type(arg):
                return True
        return False

    collection_types = (List, list, set, tuple)
    if origin is None:
        return (
            isinstance(annotation, collection_types)
            if not isinstance(annotation, type)
            else issubclass(annotation, collection_types)
        )
    else:
        return origin in collection_types  # TODO: I guess we should handle only list


def detect_collection_fields(
    args: List[FuncParam], flatten_map: Dict[str, Tuple[str, ...]]
) -> List[str]:
    """
    Django QueryDict has values that are always lists, so we need to help django ninja to understand
    better the input parameters if it's a list or a single value
    This method detects attributes that should be treated by ninja as lists and returns this list as a result
    """
    result = [i.alias or i.name for i in args if i.is_collection]

    if flatten_map:
        args_d = {arg.alias: arg for arg in args}
        for path in (p for p in flatten_map.values() if len(p) > 1):
            annotation_or_field: Any = args_d[path[0]].annotation
            for attr in path[1:]:
                if hasattr(annotation_or_field, "annotation"):
                    annotation_or_field = annotation_or_field.annotation
                annotation_or_field = next(
                    (
                        a
                        for a in annotation_or_field.model_fields.values()
                        if a.alias == attr
                    ),
                    annotation_or_field.model_fields.get(attr),
                )  # pragma: no cover

                annotation_or_field = getattr(
                    annotation_or_field, "outer_type_", annotation_or_field
                )

            # if hasattr(annotation_or_field, "annotation"):
            annotation_or_field = annotation_or_field.annotation

            if is_collection_type(annotation_or_field):
                result.append(path[-1])
    return result



import pytest
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured

from ninja import NinjaAPI, Router
from ninja.testing import TestAsyncClient, TestClient
from ninja.throttling import (
    AnonRateThrottle,
    AuthRateThrottle,
    BaseThrottle,
    SimpleRateThrottle,
    UserRateThrottle,
)


@pytest.fixture(autouse=True)
def clear_cache_for_every_case():
    cache.clear()


def test_global_throttling():
    th = AnonRateThrottle("1/s")
    set_throttle_timer(th, 0)

    api = NinjaAPI(throttle=[th])

    @api.get("/check")
    def check(request):
        return "OK"

    client = TestClient(api)

    resp = client.get("/check")
    assert resp.status_code == 200
    assert resp.content == b'"OK"'

    resp = client.get("/check")
    assert resp.status_code == 429
    assert resp.json() == {"detail": "Too many requests."}

    set_throttle_timer(th, 2)
    resp = client.get("/check")
    assert resp.status_code == 200
    assert resp.content == b'"OK"'


def test_router_throttling():
    th = AnonRateThrottle("1/s")
    set_throttle_timer(th, 0)

    api = NinjaAPI()
    router = Router()

    @router.get("/check")
    def check(request):
        return "OK"

    api.add_router("/router", router, throttle=th)

    client = TestClient(api)

    resp = client.get("/router/check")
    assert resp.status_code == 200
    assert resp.content == b'"OK"'

    resp = client.get("/router/check")
    assert resp.status_code == 429
    assert resp.json() == {"detail": "Too many requests."}


def test_router2_throttling():
    "Here we test that child router inherits the throttling from api instance"
    th = AnonRateThrottle("1/s")
    set_throttle_timer(th, 0)

    api = NinjaAPI(throttle=th)
    router = Router()

    @router.get("/check")
    def check(request):
        return "OK"

    api.add_router("/router", router)

    client = TestClient(api)

    resp = client.get("/router/check")
    assert resp.status_code == 200
    assert resp.content == b'"OK"'

    resp = client.get("/router/check")
    assert resp.status_code == 429
    assert resp.json() == {"detail": "Too many requests."}


def test_operation_throttling():
    th = AnonRateThrottle("1/s")
    set_throttle_timer(th, 0)

    api = NinjaAPI()

    @api.get("/check1", throttle=th)
    def check(request):
        return "OK"

    client = TestClient(api)

    resp = client.get("/check1")
    assert resp.status_code == 200
    assert resp.content == b'"OK"'

    resp = client.get("/check1")
    assert resp.status_code == 429
    assert resp.json() == {"detail": "Too many requests."}


@pytest.mark.asyncio
async def test_async_throttling():
    th = AnonRateThrottle("1/s")
    set_throttle_timer(th, 0)

    api = NinjaAPI(throttle=th)

    @api.get("/check-async")
    async def check(request):
        return "OK"

    client = TestAsyncClient(api)

    resp = await client.get("/check-async")
    assert resp.status_code == 200

    resp = await client.get("/check-async")
    assert resp.status_code == 429


# "Unit tests" for the throttling module

_client = TestClient(NinjaAPI())


def build_request(addr="8.8.8.8", x_forwarded_for=None):
    "Creates a mock request with the given address and optional X-Forwarded-For header."
    meta = {"REMOTE_ADDR": addr}
    if x_forwarded_for:
        meta["HTTP_X_FORWARDED_FOR"] = x_forwarded_for
    return _client._build_request("GET", "/", {}, {"META": meta})


def test_throttle_anon():
    th = AnonRateThrottle("1/s")
    set_throttle_timer(th, 0)

    request = build_request()
    request.auth = None

    assert th.allow_request(request) is True
    assert th.wait() == 1.0
    assert th.get_cache_key(request) == "throttle_anon_8.8.8.8"

    # Next should not allow as it's within the same second
    assert th.allow_request(request) is False

    # For auth request it should always allowed
    request.auth = "some"
    assert th.allow_request(request) is True
    assert th.allow_request(request) is True
    assert th.allow_request(request) is True
    assert th.get_cache_key(request) is None


def test_throttle_auth():
    th = AuthRateThrottle("1/s")
    set_throttle_timer(th, 0)

    request = build_request()
    request.auth = None

    assert th.allow_request(request) is True
    assert th.allow_request(request) is False

    request.auth = "some"
    assert th.allow_request(request) is True
    assert th.allow_request(request) is False

    set_throttle_timer(th, 2)
    assert th.allow_request(request) is True

    assert (
        th.get_cache_key(request)
        == "throttle_auth_a6b46dd0d1ae5e86cbc8f37e75ceeb6760230c1ca4ffbcb0c97b96dd7d9c464b"
    )


def test_throttle_user():
    th = UserRateThrottle("1/s")
    set_throttle_timer(th, 0)

    request = build_request()
    request.user.is_authenticated = True
    request.user.pk = 123

    assert th.allow_request(request) is True
    assert th.allow_request(request) is False

    set_throttle_timer(th, 2)
    assert th.allow_request(request) is True

    assert th.get_cache_key(request) == "throttle_user_123"

    # Not authenticated user:
    request.user.is_authenticated = False
    assert th.allow_request(request) is True
    assert th.allow_request(request) is False
    assert (
        th.get_cache_key(request) == "throttle_user_8.8.8.8"
    )  # not authenticated throttled by IP


def test_wait():
    th = AuthRateThrottle("5/m")
    set_throttle_timer(th, 0)

    request = build_request()
    request.auth = None

    for _i in range(5):
        assert th.allow_request(request) is True

    assert th.allow_request(request) is False
    assert th.wait() == 60

    set_throttle_timer(th, 30)
    assert th.allow_request(request) is False
    assert th.wait() == 30

    # Simulating cache expiration/reset
    th.history = []
    # cache.clear()
    set_throttle_timer(th, 0)
    assert th.wait() == 10  # 60s / 6 available

    # Simulating larger history
    th.history = [0] * 10
    th.now = 0
    assert th.wait() is None  # available becomes negative


def test_rate_parser():
    th = SimpleRateThrottle("1/s")
    assert th.parse_rate(None) == (None, None)
    assert th.parse_rate("1/s") == (1, 1)
    assert th.parse_rate("1/sec") == (1, 1)
    assert th.parse_rate("100/10s") == (100, 10)
    assert th.parse_rate("100/10sec") == (100, 10)
    assert th.parse_rate("100/10") == (100, 10)
    assert th.parse_rate("5/m") == (5, 60)
    assert th.parse_rate("5/min") == (5, 60)
    assert th.parse_rate("500/10m") == (500, 600)
    assert th.parse_rate("500/10min") == (500, 600)
    assert th.parse_rate("10/h") == (10, 3600)
    assert th.parse_rate("10/hour") == (10, 3600)
    assert th.parse_rate("1000/2h") == (1000, 7200)
    assert th.parse_rate("1000/2hour") == (1000, 7200)
    assert th.parse_rate("100/d") == (100, 86400)
    assert th.parse_rate("100/day") == (100, 86400)
    assert th.parse_rate("10_000/7d") == (10000, 86400 * 7)
    assert th.parse_rate("10_000/7day") == (10000, 86400 * 7)

    with pytest.raises(ValueError):
        th.parse_rate("42")


def test_proxy_throttle():
    from ninja.conf import settings

    settings.NUM_PROXIES = 0  # instead of None

    th = SimpleRateThrottle("1/s")
    request = build_request(x_forwarded_for=None)
    assert th.get_ident(request) == "8.8.8.8"

    settings.NUM_PROXIES = 0
    request = build_request(x_forwarded_for="8.8.8.8,127.0.0.1")
    assert th.get_ident(request) == "8.8.8.8"

    settings.NUM_PROXIES = 1
    assert th.get_ident(request) == "127.0.0.1"

    settings.NUM_PROXIES = None


def test_base_classes():
    base = BaseThrottle()
    with pytest.raises(NotImplementedError):
        base.allow_request(build_request())
    assert base.wait() is None

    sample = SimpleRateThrottle("1/s")
    with pytest.raises(NotImplementedError):
        sample.allow_request(build_request())

    throttle = AnonRateThrottle()
    with pytest.raises(ImproperlyConfigured):
        throttle.scope = None
        throttle.get_rate()

    sample_scope2 = SimpleRateThrottle("1/s")
    sample_scope2.scope = "scope2"
    with pytest.raises(ImproperlyConfigured):
        sample_scope2.get_rate()


def set_throttle_timer(throttle: BaseThrottle, value: int):
    """
    Explicitly set the timer, overriding time.time()
    """
    throttle.timer = lambda: value



import sys
from sys import version_info
from typing import Any, List, Union
from unittest.mock import Mock
from uuid import uuid4

import pydantic
import pytest
from django.contrib.admin.views.decorators import staff_member_required
from django.db import models
from django.test import Client, override_settings
from typing_extensions import Annotated

from ninja import (
    Body,
    Field,
    File,
    Form,
    ModelSchema,
    NinjaAPI,
    P,
    PathEx,
    Query,
    Schema,
    UploadedFile,
)
from ninja.openapi.urls import get_openapi_urls
from ninja.pagination import PaginationBase, paginate
from ninja.renderers import JSONRenderer

api = NinjaAPI()


class Payload(Schema):
    i: int
    f: float


class TypeA(Schema):
    a: str


class TypeB(Schema):
    b: str


AnnotatedStr = Annotated[
    str,
    pydantic.WithJsonSchema({
        "type": "string",
        "format": "custom-format",
        "example": "example_string",
    }),
]


def to_camel(string: str) -> str:
    words = string.split("_")
    return words[0].lower() + "".join(word.capitalize() for word in words[1:])


class Response(Schema):
    model_config = pydantic.ConfigDict(alias_generator=to_camel, populate_by_name=True)
    i: int
    f: float = Field(..., title="f title", description="f desc")


@api.post("/test", response=Response)
def method(request, data: Payload):
    return data.dict()


@api.post("/test-alias", response=Response, by_alias=True)
def method_alias(request, data: Payload):
    return data.dict()


@api.post("/test_list", response=List[Response])
def method_list_response(request, data: List[Payload]):
    return []


@api.post("/test-body", response=Response)
def method_body(request, i: int = Body(...), f: float = Body(...)):
    return dict(i=i, f=f)


@api.post("/test-body-schema", response=Response)
def method_body_schema(request, data: Payload):
    return dict(i=data.i, f=data.f)


@api.get("/test-path/{int:i}/{f}", response=Response)
def method_path(
    request,
    i: int,
    f: float,
):
    return dict(i=i, f=f)


# This definition is only possible in Python 3.9+
# TODO: Drop this condition once support for <= 3.8 is dropped
if version_info >= (3, 9):

    @api.get("/test-pathex/{path_ex}", response=AnnotatedStr)
    def method_pathex(
        request,
        path_ex: PathEx[
            AnnotatedStr,
            P(description="path_ex description"),
        ],
    ):
        return path_ex

else:
    with pytest.raises(NotImplementedError, match="3.9+"):

        @api.get("/test-pathex/{path_ex}", response=AnnotatedStr)
        def method_pathex(
            request,
            path_ex: PathEx[
                AnnotatedStr,
                P(description="path_ex description"),
            ],
        ):
            return path_ex


@api.post("/test-form", response=Response)
def method_form(request, data: Payload = Form(...)):
    return dict(i=data.i, f=data.f)


@api.post("/test-form-single", response=Response)
def method_form_single(request, data: float = Form(...)):
    return dict(i=int(data), f=data)


@api.post("/test-form-body", response=Response)
def method_form_body(request, i: int = Form(10), s: str = Body("10")):
    return dict(i=i, s=s)


@api.post("/test-form-file", response=Response)
def method_form_file(request, files: List[UploadedFile], data: Payload = Form(...)):
    return dict(i=data.i, f=data.f)


@api.post("/test-body-file", response=Response)
def method_body_file(
    request,
    files: List[UploadedFile],
    body: Payload = Body(...),
):
    return dict(i=body.i, f=body.f)


@api.post("/test-union-type", response=Response)
def method_union_payload(request, data: Union[TypeA, TypeB]):
    return dict(i=data.i, f=data.f)


@api.post("/test-union-type-with-simple", response=Response)
def method_union_payload_and_simple(request, data: Union[int, TypeB]):
    return data.dict()


if sys.version_info >= (3, 10):
    # This requires Python 3.10 or higher (PEP 604), so we're using eval to
    # conditionally make it available
    @api.post("/test-new-union-type", response=Response)
    def method_new_union_payload(request, data: "TypeA | TypeB"):
        return dict(i=data.i, f=data.f)


@api.post(
    "/test-title-description/",
    tags=["a-tag"],
    summary="Best API Ever",
    response=Response,
)
def method_test_title_description(
    request,
    param1: int = Query(..., title="param 1 title"),
    param2: str = Query("A Default", description="param 2 desc"),
    file: UploadedFile = File(..., description="file param desc"),
):
    return dict(i=param1, f=param2)


@api.post("/test-deprecated-example-examples/")
def method_test_deprecated_example_examples(
    request,
    param1: int = Query(None, deprecated=True),
    param2: str = Query(..., example="Example Value"),
    param3: str = Query(
        ...,
        max_length=5,
        examples={
            "normal": {
                "summary": "A normal example",
                "description": "A **normal** string works correctly.",
                "value": "Foo",
            },
            "invalid": {
                "summary": "Invalid data is rejected with an error",
                "value": "MoreThan5Length",
            },
        },
    ),
    param4: int = Query(None, deprecated=True, include_in_schema=False),
):
    return dict(i=param2, f=param3)


def test_schema_views(client: Client):
    assert client.get("/api/").status_code == 404
    assert client.get("/api/docs").status_code == 200
    assert client.get("/api/openapi.json").status_code == 200


def test_schema_views_no_INSTALLED_APPS(client: Client):
    "Making sure that cdn and included js works fine"
    from django.conf import settings

    # removing ninja from settings:
    INSTALLED_APPS = [i for i in settings.INSTALLED_APPS if i != "ninja"]

    @override_settings(INSTALLED_APPS=INSTALLED_APPS)
    def call_docs():
        assert client.get("/api/docs").status_code == 200

    call_docs()


@pytest.fixture(scope="session")
def schema():
    return api.get_openapi_schema()


def test_schema(schema):
    method = schema["paths"]["/api/test"]["post"]

    assert method["requestBody"] == {
        "content": {
            "application/json": {"schema": {"$ref": "#/components/schemas/Payload"}}
        },
        "required": True,
    }
    assert method["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
            "description": "OK",
        }
    }
    assert schema.schemas == {
        "Response": {
            "title": "Response",
            "type": "object",
            "properties": {
                "i": {"title": "I", "type": "integer"},
                "f": {"description": "f desc", "title": "f title", "type": "number"},
            },
            "required": ["i", "f"],
        },
        "Payload": {
            "title": "Payload",
            "type": "object",
            "properties": {
                "i": {"title": "I", "type": "integer"},
                "f": {"title": "F", "type": "number"},
            },
            "required": ["i", "f"],
        },
        "TypeA": {
            "properties": {
                "a": {"title": "A", "type": "string"},
            },
            "required": ["a"],
            "title": "TypeA",
            "type": "object",
        },
        "TypeB": {
            "properties": {
                "b": {"title": "B", "type": "string"},
            },
            "required": ["b"],
            "title": "TypeB",
            "type": "object",
        },
    }


def test_schema_alias(schema):
    method = schema["paths"]["/api/test-alias"]["post"]

    assert method["requestBody"] == {
        "content": {
            "application/json": {"schema": {"$ref": "#/components/schemas/Payload"}}
        },
        "required": True,
    }
    assert method["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
            "description": "OK",
        }
    }
    # ::TODO:: this is currently broken if not all responses for same schema use the same by_alias
    """
    assert schema.schemas == {
        "Response": {
            "title": "Response",
            "type": "object",
            "properties": {
                "I": {"title": "I", "type": "integer"},
                "F": {"title": "F", "type": "number"},
            },
            "required": ["i", "f"],
        },
        "Payload": {
            "title": "Payload",
            "type": "object",
            "properties": {
                "i": {"title": "I", "type": "integer"},
                "f": {"title": "F", "type": "number"},
            },
            "required": ["i", "f"],
        },
    }
    """


def test_schema_list(schema):
    method_list = schema["paths"]["/api/test_list"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "items": {"$ref": "#/components/schemas/Payload"},
                    "title": "Data",
                    "type": "array",
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {
                        "items": {"$ref": "#/components/schemas/Response"},
                        "title": "Response",
                        "type": "array",
                    }
                }
            },
            "description": "OK",
        }
    }

    assert schema["components"]["schemas"] == {
        "Payload": {
            "properties": {
                "f": {"title": "F", "type": "number"},
                "i": {"title": "I", "type": "integer"},
            },
            "required": ["i", "f"],
            "title": "Payload",
            "type": "object",
        },
        "TypeA": {
            "properties": {
                "a": {"title": "A", "type": "string"},
            },
            "required": ["a"],
            "title": "TypeA",
            "type": "object",
        },
        "TypeB": {
            "properties": {
                "b": {"title": "B", "type": "string"},
            },
            "required": ["b"],
            "title": "TypeB",
            "type": "object",
        },
        "Response": {
            "properties": {
                "f": {"description": "f desc", "title": "f title", "type": "number"},
                "i": {"title": "I", "type": "integer"},
            },
            "required": ["i", "f"],
            "title": "Response",
            "type": "object",
        },
    }


def test_schema_body(schema):
    method_list = schema["paths"]["/api/test-body"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "properties": {
                        "f": {"title": "F", "type": "number"},
                        "i": {"title": "I", "type": "integer"},
                    },
                    "required": ["i", "f"],
                    "title": "BodyParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
            "description": "OK",
        }
    }


def test_schema_body_schema(schema):
    method_list = schema["paths"]["/api/test-body-schema"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "application/json": {"schema": {"$ref": "#/components/schemas/Payload"}},
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
            "description": "OK",
        }
    }


def test_schema_path(schema):
    method_list = schema["paths"]["/api/test-path/{i}/{f}"]["get"]

    assert "requestBody" not in method_list

    assert method_list["parameters"] == [
        {
            "in": "path",
            "name": "i",
            "schema": {"title": "I", "type": "integer"},
            "required": True,
        },
        {
            "in": "path",
            "name": "f",
            "schema": {"title": "F", "type": "number"},
            "required": True,
        },
    ]

    assert method_list["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"},
                },
            },
            "description": "OK",
        }
    }


@pytest.mark.skipif(
    version_info < (3, 9),
    reason="requires py3.9+ for Annotated[] at the route definition site",
)
def test_schema_pathex(schema):
    method_list = schema["paths"]["/api/test-pathex/{path_ex}"]["get"]

    assert "requestBody" not in method_list

    assert method_list["parameters"] == [
        {
            "in": "path",
            "name": "path_ex",
            "schema": {
                "title": "Path Ex",
                "type": "string",
                "format": "custom-format",
                "description": "path_ex description",
                "example": "example_string",
            },
            "required": True,
            "example": "example_string",
            "description": "path_ex description",
        },
    ]

    assert method_list["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {
                        "example": "example_string",
                        "format": "custom-format",
                        "title": "Response",
                        "type": "string",
                    },
                },
            },
            "description": "OK",
        }
    }


def test_schema_form(schema):
    method_list = schema["paths"]["/api/test-form"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "application/x-www-form-urlencoded": {
                "schema": {
                    "title": "FormParams",
                    "type": "object",
                    "properties": {
                        "i": {"title": "I", "type": "integer"},
                        "f": {"title": "F", "type": "number"},
                    },
                    "required": ["i", "f"],
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "description": "OK",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
        }
    }


def test_schema_single(schema):
    method_list = schema["paths"]["/api/test-form-single"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "application/x-www-form-urlencoded": {
                "schema": {
                    "properties": {"data": {"title": "Data", "type": "number"}},
                    "required": ["data"],
                    "title": "FormParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "description": "OK",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
        }
    }


def test_schema_form_body(schema):
    method_list = schema["paths"]["/api/test-form-body"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "multipart/form-data": {
                "schema": {
                    "properties": {
                        "i": {"default": 10, "title": "I", "type": "integer"},
                        "s": {"default": "10", "title": "S", "type": "string"},
                    },
                    "title": "MultiPartBodyParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "description": "OK",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
        }
    }


def test_schema_form_file(schema):
    method_list = schema["paths"]["/api/test-form-file"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "multipart/form-data": {
                "schema": {
                    "properties": {
                        "files": {
                            "items": {"format": "binary", "type": "string"},
                            "title": "Files",
                            "type": "array",
                        },
                        "i": {"title": "I", "type": "integer"},
                        "f": {"title": "F", "type": "number"},
                    },
                    "required": ["files", "i", "f"],
                    "title": "MultiPartBodyParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "description": "OK",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
        }
    }


def test_schema_body_file(schema):
    method_list = schema["paths"]["/api/test-body-file"]["post"]

    assert method_list["requestBody"] == {
        "content": {
            "multipart/form-data": {
                "schema": {
                    "properties": {
                        "body": {"$ref": "#/components/schemas/Payload"},
                        "files": {
                            "items": {"format": "binary", "type": "string"},
                            "title": "Files",
                            "type": "array",
                        },
                    },
                    "required": ["files", "body"],
                    "title": "MultiPartBodyParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }
    assert method_list["responses"] == {
        200: {
            "description": "OK",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
        }
    }


def test_schema_title_description(schema):
    method_list = schema["paths"]["/api/test-title-description/"]["post"]

    assert method_list["summary"] == "Best API Ever"
    assert method_list["tags"] == ["a-tag"]

    assert method_list["requestBody"] == {
        "content": {
            "multipart/form-data": {
                "schema": {
                    "properties": {
                        "file": {
                            "description": "file param desc",
                            "format": "binary",
                            "title": "File",
                            "type": "string",
                        }
                    },
                    "required": ["file"],
                    "title": "FileParams",
                    "type": "object",
                }
            }
        },
        "required": True,
    }

    assert method_list["parameters"] == [
        {
            "in": "query",
            "name": "param1",
            "required": True,
            "schema": {"title": "param 1 title", "type": "integer"},
        },
        {
            "in": "query",
            "name": "param2",
            "description": "param 2 desc",
            "required": False,
            "schema": {
                "default": "A Default",
                "description": "param 2 desc",
                "title": "Param2",
                "type": "string",
            },
        },
    ]

    assert method_list["responses"] == {
        200: {
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/Response"}
                }
            },
            "description": "OK",
        }
    }


def test_schema_deprecated_example_examples(schema):
    method_list = schema["paths"]["/api/test-deprecated-example-examples/"]["post"]

    assert method_list["parameters"] == [
        {
            "deprecated": True,
            "in": "query",
            "name": "param1",
            "required": False,
            "schema": {"title": "Param1", "type": "integer", "deprecated": True},
        },
        {
            "in": "query",
            "name": "param2",
            "required": True,
            "schema": {"title": "Param2", "type": "string", "example": "Example Value"},
            "example": "Example Value",
        },
        {
            "in": "query",
            "name": "param3",
            "required": True,
            "schema": {
                "maxLength": 5,
                "title": "Param3",
                "type": "string",
                "examples": {
                    "invalid": {
                        "summary": "Invalid data is rejected with an error",
                        "value": "MoreThan5Length",
                    },
                    "normal": {
                        "description": "A **normal** string works correctly.",
                        "summary": "A normal example",
                        "value": "Foo",
                    },
                },
            },
            "examples": {
                "invalid": {
                    "summary": "Invalid data is rejected with an error",
                    "value": "MoreThan5Length",
                },
                "normal": {
                    "description": "A **normal** string works correctly.",
                    "summary": "A normal example",
                    "value": "Foo",
                },
            },
        },
    ]

    assert method_list["responses"] == {
        200: {
            "description": "OK",
        }
    }


def test_union_payload_type(schema):
    method = schema["paths"]["/api/test-union-type"]["post"]

    assert method["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "anyOf": [
                        {"$ref": "#/components/schemas/TypeA"},
                        {"$ref": "#/components/schemas/TypeB"},
                    ],
                    "title": "Data",
                }
            }
        },
        "required": True,
    }


def test_union_payload_simple(schema):
    method = schema["paths"]["/api/test-union-type-with-simple"]["post"]

    print(method["requestBody"])
    assert method["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "title": "Data",
                    "anyOf": [
                        {"type": "integer"},
                        {"$ref": "#/components/schemas/TypeB"},
                    ],
                }
            }
        },
        "required": True,
    }


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="requires Python 3.10 or higher (PEP 604)",
)
def test_new_union_payload_type(schema):
    method = schema["paths"]["/api/test-new-union-type"]["post"]

    assert method["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "anyOf": [
                        {"$ref": "#/components/schemas/TypeA"},
                        {"$ref": "#/components/schemas/TypeB"},
                    ],
                    "title": "Data",
                }
            }
        },
        "required": True,
    }


def test_get_openapi_urls():
    api = NinjaAPI(openapi_url=None)
    paths = get_openapi_urls(api)
    assert len(paths) == 0

    api = NinjaAPI(docs_url=None)
    paths = get_openapi_urls(api)
    assert len(paths) == 1

    api = NinjaAPI(openapi_url="/path", docs_url="/path")
    with pytest.raises(
        AssertionError, match="Please use different urls for openapi_url and docs_url"
    ):
        get_openapi_urls(api)


def test_unique_operation_ids(capsys):
    api = NinjaAPI()

    @api.get("/1")
    def same_name(request):
        pass

    @api.get("/2")  # noqa: F811
    def same_name(request):  # noqa: F811
        pass

    api.get_openapi_schema()
    captured = capsys.readouterr()
    assert '"test_openapi_schema_same_name" is already used ' in captured.out


def test_docs_decorator():
    api = NinjaAPI(docs_decorator=staff_member_required)

    paths = get_openapi_urls(api)
    assert len(paths) == 2
    for ptrn in paths:
        request = Mock(user=Mock(is_staff=True))
        result = ptrn.callback(request)
        assert result.status_code == 200

        request = Mock(user=Mock(is_staff=False))
        request.build_absolute_uri = lambda: "http://example.com"
        result = ptrn.callback(request)
        assert result.status_code == 302


class TestRenderer(JSONRenderer):
    media_type = "custom/type"


def test_renderer_media_type():
    api = NinjaAPI(renderer=TestRenderer)

    @api.get("/1", response=TypeA)
    def same_name(
        request,
    ):
        pass

    schema = api.get_openapi_schema()
    method = schema["paths"]["/api/1"]["get"]
    assert method["responses"] == {
        200: {
            "content": {
                "custom/type": {"schema": {"$ref": "#/components/schemas/TypeA"}}
            },
            "description": "OK",
        }
    }


def test_all_paths_rendered():
    api = NinjaAPI(renderer=TestRenderer)

    @api.post("/1")
    def some_name_create(
        request,
    ):
        pass

    @api.get("/1")
    def some_name_list(
        request,
    ):
        pass

    @api.get("/1/{param}")
    def some_name_get_one(request, param: int):
        pass

    @api.delete("/1/{param}")
    def some_name_delete(request, param: int):
        pass

    schema = api.get_openapi_schema()

    expected_result = {"/api/1": ["post", "get"], "/api/1/{param}": ["get", "delete"]}
    result = {p: list(schema["paths"][p].keys()) for p in schema["paths"].keys()}
    assert expected_result == result


def test_all_paths_typed_params_rendered():
    api = NinjaAPI(renderer=TestRenderer)

    @api.post("/1")
    def some_name_create(
        request,
    ):
        pass

    @api.get("/1")
    def some_name_list(
        request,
    ):
        pass

    @api.get("/1/{int:param}")
    def some_name_get_one(request, param: int):
        pass

    @api.delete("/1/{str:param}")
    def some_name_delete(request, param: str):
        pass

    schema = api.get_openapi_schema()

    expected_result = {"/api/1": ["post", "get"], "/api/1/{param}": ["get", "delete"]}
    result = {p: list(schema["paths"][p].keys()) for p in schema["paths"].keys()}
    assert expected_result == result


def test_no_default_for_custom_items_attribute():
    api = NinjaAPI(renderer=TestRenderer)

    class EmployeeOut(Schema):
        id: int
        first_name: str
        last_name: str

    class CustomPagination(PaginationBase):
        class Output(Schema):
            data: List[Any]  # `items` is a default attribute
            detail: str
            total: int

        items_attribute: str = "data"

        def paginate_queryset(self, queryset, pagination, **params):
            pass

    @api.get(
        "/employees",
        auth=["OAuth"],
        response=List[EmployeeOut],
    )
    @paginate(CustomPagination)
    def get_employees(request):
        pass

    schema = api.get_openapi_schema()

    paged_employee_out = schema["components"]["schemas"]["PagedEmployeeOut"]
    # a default value shouldn't be specified automatically
    assert "default" not in paged_employee_out["properties"]["data"]


def test_by_alias_uses_serialization_alias_simple():
    """Test the serialization_alias on the Field is used when by_alias=True is set on the route"""
    api = NinjaAPI()

    class PersonOut(Schema):
        uuid: str = Field(..., serialization_alias="id")
        name: str = Field(..., serialization_alias="fullName")

    @api.get("/person", response={200: PersonOut}, by_alias=True)
    def get_user(request):
        return {"uuid": uuid4(), "fullname": "John Snow"}

    schema = api.get_openapi_schema()
    user_alias_schema = schema["components"]["schemas"]["PersonOut"]
    assert user_alias_schema == {
        "title": "PersonOut",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "type": "string"},
            "fullName": {"type": "string", "title": "Fullname"},
        },
        "required": ["id", "fullName"],
    }


def test_by_alias_uses_validation_alias_simple():
    """Test the serialization_alias on the Field is used when by_alias=True is set on the route"""
    api = NinjaAPI()

    class PersonIn(Schema):
        uuid: str = Field(..., validation_alias="id")
        name: str = Field(..., validation_alias="fullName")

    @api.get("/person", by_alias=True)
    def get_user(request, param: PersonIn):
        return {"uuid": uuid4(), "fullname": "John Snow"}

    schema = api.get_openapi_schema()
    user_alias_schema = schema["components"]["schemas"]["PersonIn"]
    assert user_alias_schema == {
        "title": "PersonIn",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "type": "string"},
            "fullName": {"type": "string", "title": "Fullname"},
        },
        "required": ["id", "fullName"],
    }


@pytest.mark.django_db
def test_by_alias_uses_serialization_alias_model():
    """Test the serialization_alias on the Field is used when by_alias=True is set on the route"""
    api = NinjaAPI()

    class Person(models.Model):
        uuid = models.CharField(
            max_length=36, unique=True, default=lambda: str(uuid4()), editable=False
        )
        created = models.DateTimeField(auto_now_add=True)

        class Meta:
            app_label = "tests"

    class PersonModelOut(ModelSchema):
        uuid: str = Field(..., serialization_alias="id")

        class Meta:
            model = Person
            fields = ["created"]

    @api.get("/person", response={200: PersonModelOut}, by_alias=True)
    def get_user(request):
        return Person(fullname="John Snow")

    schema = api.get_openapi_schema()
    user_alias_schema = schema["components"]["schemas"]["PersonModelOut"]
    assert user_alias_schema == {
        "title": "PersonModelOut",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "type": "string"},
            "created": {"type": "string", "format": "date-time", "title": "Created"},
        },
        "required": ["id", "created"],
    }



from unittest import mock

import pytest
from pydantic import model_serializer

from ninja import Router, Schema
from ninja.schema import pydantic_version
from ninja.testing import TestClient


def api_endpoint_test(request):
    return {
        "test1": "foo",
        "test2": "bar",
    }


@pytest.mark.skipif(
    pydantic_version < [2, 7],
    reason="Serialization context was introduced in Pydantic 2.7",
)
def test_request_is_passed_in_context_when_supported():
    class SchemaWithCustomSerializer(Schema):
        test1: str
        test2: str

        @model_serializer(mode="wrap")
        def ser_model(self, handler, info):
            assert "request" in info.context
            assert info.context["request"].path == "/test"  # check it is HttRequest
            assert "response_status" in info.context

            return handler(self)

    router = Router()
    router.add_api_operation(
        "/test", ["GET"], api_endpoint_test, response=SchemaWithCustomSerializer
    )

    TestClient(router).get("/test")


@pytest.mark.parametrize(
    ["pydantic_version"],
    [
        [[2, 0]],
        [[2, 4]],
        [[2, 6]],
    ],
)
def test_no_serialisation_context_used_when_no_supported(pydantic_version):
    class SchemaWithCustomSerializer(Schema):
        test1: str
        test2: str

        @model_serializer(mode="wrap")
        def ser_model(self, handler, info):
            if hasattr(info, "context"):
                # an actually newer Pydantic, but pydantic_version is still mocked, so no context is expected
                assert info.context is None

            return handler(self)

    with mock.patch("ninja.operation.pydantic_version", pydantic_version):
        router = Router()
        router.add_api_operation(
            "/test", ["GET"], api_endpoint_test, response=SchemaWithCustomSerializer
        )

        resp_json = TestClient(router).get("/test").json()

        assert resp_json == {
            "test1": "foo",
            "test2": "bar",
        }



import pytest
from django.http import Http404

from ninja import NinjaAPI, Schema
from ninja.testing import TestAsyncClient, TestClient

api = NinjaAPI()


class CustomException(Exception):
    pass


@api.exception_handler(CustomException)
def on_custom_error(request, exc):
    return api.create_response(request, {"custom": True}, status=422)


class Payload(Schema):
    test: int


@api.post("/error/{code}")
def err_thrower(request, code: str, payload: Payload = None):
    if code == "base":
        raise RuntimeError("test")
    if code == "404":
        raise Http404("test")
    if code == "custom":
        raise CustomException("test")


client = TestClient(api)


def test_default_handler(settings):
    settings.DEBUG = True

    response = client.post("/error/base")
    assert response.status_code == 500
    assert b"RuntimeError: test" in response.content

    response = client.post("/error/404")
    assert response.status_code == 404
    assert response.json() == {"detail": "Not Found: test"}

    response = client.post("/error/custom", body="invalid_json")
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Cannot parse request body (Expecting value: line 1 column 1 (char 0))",
    }

    settings.DEBUG = False
    with pytest.raises(RuntimeError):
        response = client.post("/error/base")

    response = client.post("/error/custom", body="invalid_json")
    assert response.status_code == 400
    assert response.json() == {"detail": "Cannot parse request body"}


@pytest.mark.parametrize(
    "route,status_code,json",
    [
        ("/error/404", 404, {"detail": "Not Found"}),
        ("/error/custom", 422, {"custom": True}),
    ],
)
def test_exceptions(route, status_code, json):
    response = client.post(route)
    assert response.status_code == status_code
    assert response.json() == json


@pytest.mark.asyncio
async def test_asyncio_exceptions():
    api = NinjaAPI()

    @api.get("/error")
    async def thrower(request):
        raise Http404("test")

    client = TestAsyncClient(api)
    response = await client.get("/error")
    assert response.status_code == 404


def test_no_handlers():
    api = NinjaAPI()
    api._exception_handlers = {}

    @api.get("/error")
    def thrower(request):
        raise RuntimeError("test")

    client = TestClient(api)

    with pytest.raises(RuntimeError):
        client.get("/error")



import os
import sys
from pathlib import Path

ROOT = Path(__file__).parent.parent.resolve()

sys.path.insert(0, str(ROOT))
sys.path.insert(0, str(ROOT / "tests/demo_project"))

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "demo.settings")

import django  # noqa

django.setup()


def pytest_generate_tests(metafunc):
    os.environ["NINJA_SKIP_REGISTRY"] = "yes"



import pytest
from django.urls import reverse


@pytest.mark.parametrize(
    "view_name, path",
    [
        ("foobar", "/api/v3/foobar"),
        ("post_foobar", "/api/v3/foobar"),
        ("foobar_put", "/api/v3/foobar"),
    ],
)
def test_reverse(view_name, path):
    assert reverse(f"api-3.0.0:{view_name}") == path



from ninja.security import APIKeyQuery
from someapp.models import Client


class ApiKey(APIKeyQuery):
    param_name = "api_key"

    def authenticate(self, request, key):
        try:
            return Client.objects.get(key=key)
        except Client.DoesNotExist:
            pass


api_key = ApiKey()


@api.get("/apikey", auth=api_key)
def apikey(request):
    assert isinstance(request.auth, Client)
    return f"Hello {request.auth}"



from ninja import NinjaAPI, Router
from ninja.testing import TestClient

router = Router()


@router.get("/")
def op(request):
    return True


def test_add_router_with_string_path():
    main_router = Router()
    main_router.add_router("sub", "tests.test_router_add_router.router")

    api = NinjaAPI()
    api.add_router("main", main_router)

    client = TestClient(api)

    assert client.get("/main/sub/").status_code == 200



import typing
from sys import version_info

import pytest

from ninja.signature.details import is_collection_type


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        pytest.param(typing.List, True, id="true_for_typing_List"),
        pytest.param(list, True, id="true_for_native_list"),
        pytest.param(typing.Set, True, id="true_for_typing_Set"),
        pytest.param(set, True, id="true_for_native_set"),
        pytest.param(typing.Tuple, True, id="true_for_typing_Tuple"),
        pytest.param(tuple, True, id="true_for_native_tuple"),
        pytest.param(
            typing.Optional[typing.List[str]], True, id="true_for_optional_list"
        ),
        pytest.param(
            type("Custom", (), {}),
            False,
            id="false_for_custom_type_without_typing_origin",
        ),
        pytest.param(
            object(), False, id="false_for_custom_instance_without_typing_origin"
        ),
        pytest.param(
            typing.NewType("SomethingNew", str),
            False,
            id="false_for_instance_without_typing_origin",
        ),
        # Can't mark with `pytest.mark.skipif` since we'd attempt to instantiate the
        # parameterized value/type(e.g. `list[int]`). Which only works with Python >= 3.9)
        *(
            (
                pytest.param(list[int], True, id="true_for_parameterized_native_list"),
                pytest.param(set[int], True, id="true_for_parameterized_native_set"),
                pytest.param(
                    tuple[int], True, id="true_for_parameterized_native_tuple"
                ),
            )
            # TODO: Remove conditional once support for <=3.8 is dropped
            if version_info >= (3, 9)
            else ()
        ),
    ],
)
def test_is_collection_type_returns(annotation: typing.Any, expected: bool):
    assert is_collection_type(annotation) is expected



import json
from typing import Any, Mapping, Optional, Type

from django.http import HttpRequest

from ninja.responses import NinjaJSONEncoder

__all__ = ["BaseRenderer", "JSONRenderer"]


class BaseRenderer:
    media_type: Optional[str] = None
    charset: str = "utf-8"

    def render(self, request: HttpRequest, data: Any, *, response_status: int) -> Any:
        raise NotImplementedError("Please implement .render() method")


class JSONRenderer(BaseRenderer):
    media_type = "application/json"
    encoder_class: Type[json.JSONEncoder] = NinjaJSONEncoder
    json_dumps_params: Mapping[str, Any] = {}

    def render(self, request: HttpRequest, data: Any, *, response_status: int) -> Any:
        return json.dumps(data, cls=self.encoder_class, **self.json_dumps_params)



from typing import Union

__all__ = ["UNION_TYPES"]


# python3.10+ syntax of creating a union or optional type (with str | int)
# UNION_TYPES allows to check both universes if types are a union
try:
    from types import UnionType

    UNION_TYPES = (Union, UnionType)
except ImportError:
    UNION_TYPES = (Union,)



from ninja import NinjaAPI


class TestServer:
    def test_server_basic(self):
        server = {"url": "http://example.com"}
        api = NinjaAPI(servers=[server])
        schema = api.get_openapi_schema()

        schema_server = schema["servers"]
        assert schema_server == [server]

    def test_server_with_description(self):
        server = {
            "url": "http://example.com",
            "description": "this is the example server",
        }
        api = NinjaAPI(servers=[server])
        schema = api.get_openapi_schema()

        schema_server = schema["servers"]
        assert schema_server == [server]

    def test_multiple_servers_with_description(self):
        server_1 = {
            "url": "http://example1.com",
            "description": "this is the example server 1",
        }
        server_2 = {
            "url": "http://example2.com",
            "description": "this is the example server 2",
        }
        servers = [server_1, server_2]
        api = NinjaAPI(servers=servers)
        schema = api.get_openapi_schema()

        schema_server = schema["servers"]
        assert schema_server == servers



import asyncio
from typing import Any, List

import django
import pytest
from django.db.models import QuerySet
from someapp.models import Category

from ninja import NinjaAPI, Schema
from ninja.errors import ConfigError
from ninja.pagination import (
    AsyncPaginationBase,
    PageNumberPagination,
    PaginationBase,
    paginate,
)
from ninja.testing import TestAsyncClient

api = NinjaAPI()

ITEMS = list(range(100))


class NoAsyncPagination(PaginationBase):
    # only offset param, defaults to 5 per page
    class Input(Schema):
        skip: int

    class Output(Schema):
        items: List[Any]
        count: str
        skip: int

    def paginate_queryset(self, items, pagination: Input, **params):
        skip = pagination.skip
        return {
            "items": items[skip : skip + 5],
            "count": "many",
            "skip": skip,
        }


class AsyncNoOutputPagination(AsyncPaginationBase):
    # Outputs items without count attribute
    class Input(Schema):
        skip: int

    Output = None

    def paginate_queryset(self, items, pagination: Input, **params):
        skip = pagination.skip
        return items[skip : skip + 5]

    async def apaginate_queryset(self, items, pagination: Input, **params):
        await asyncio.sleep(0)
        skip = pagination.skip
        return items[skip : skip + 5]

    async def _items_count(self, queryset: QuerySet) -> int:
        try:
            # forcing to find queryset.count instead of list.count:
            return queryset.all().count()
        except AttributeError:
            await asyncio.sleep(0)
            return len(queryset)


class AsyncNoPagination(AsyncPaginationBase):
    """
    Async pagination class that returns all records without slicing.
    Does not define its own Input - uses empty PaginationBase.Input.
    This tests the bug fix from https://github.com/vitalik/django-ninja/issues/1564
    """

    async def apaginate_queryset(
        self, items, pagination: PaginationBase.Input, **params
    ):
        await asyncio.sleep(0)
        return {
            "count": await self._aitems_count(items),
            "items": items,
        }

    def paginate_queryset(self, items, pagination: PaginationBase.Input, **params):
        return {
            "count": self._items_count(items),
            "items": items,
        }


@pytest.mark.asyncio
async def test_async_config_error():
    api = NinjaAPI()

    with pytest.raises(
        ConfigError, match="Pagination class not configured for async requests"
    ):

        @api.get("/items_async_undefined", response=List[int])
        @paginate(NoAsyncPagination)
        async def items_async_undefined(request, **kwargs):
            return ITEMS


@pytest.mark.asyncio
async def test_async_custom_pagination():
    api = NinjaAPI()

    @api.get("/items_async", response=List[int])
    @paginate(AsyncNoOutputPagination)
    async def items_async(request):
        return ITEMS

    client = TestAsyncClient(api)

    response = await client.get("/items_async?skip=10")
    assert response.json() == [10, 11, 12, 13, 14]


@pytest.mark.asyncio
async def test_async_default():
    api = NinjaAPI()

    @api.get("/items_default", response=List[int])
    @paginate  # WITHOUT brackets (should use default pagination)
    async def items_default(request, someparam: int = 0, **kwargs):
        await asyncio.sleep(0)
        return ITEMS

    client = TestAsyncClient(api)

    response = await client.get("/items_default?limit=10")
    assert response.json() == {"items": ITEMS[:10], "count": 100}


@pytest.mark.asyncio
async def test_async_page_number():
    api = NinjaAPI()

    @api.get("/items_page_number", response=List[Any])
    @paginate(PageNumberPagination, page_size=10, pass_parameter="page_info")
    async def items_page_number(request, **kwargs):
        return ITEMS + [kwargs["page_info"]]

    client = TestAsyncClient(api)

    response = await client.get("/items_page_number?page=11")
    assert response.json() == {"items": [{"page": 11, "page_size": None}], "count": 101}


@pytest.mark.skipif(django.VERSION[:2] < (5, 0), reason="Requires Django 5.0+")
@pytest.mark.django_db
@pytest.mark.asyncio
async def test_test_async_pagination():
    await Category.objects.acreate(title="cat1")
    await Category.objects.acreate(title="cat2")
    assert await Category.objects.acount() == 2

    class CatSchema(Schema):
        title: str

    api = NinjaAPI()

    @api.get("/cats", response=list[CatSchema])
    @paginate
    async def cats_paginated_limit_offset(request):
        return Category.objects.order_by("id")

    @api.get("/cats-pages", response=list[CatSchema])
    @paginate(PageNumberPagination)
    async def cats_paginated_page_number(request):
        return Category.objects.order_by("id")

    client = TestAsyncClient(api)

    response = await client.get("/cats")
    assert response.status_code == 200
    assert response.json() == {
        "items": [{"title": "cat1"}, {"title": "cat2"}],
        "count": 2,
    }

    response = await client.get("/cats?offset=1")
    assert response.status_code == 200
    print(response.json())
    assert response.json() == {
        "items": [{"title": "cat2"}],
        "count": 2,
    }

    response = await client.get("/cats-pages")
    assert response.status_code == 200
    assert response.json() == {
        "items": [{"title": "cat1"}, {"title": "cat2"}],
        "count": 2,
    }

    response = await client.get("/cats-pages?page=1")
    assert response.status_code == 200
    assert response.json() == {
        "items": [{"title": "cat1"}, {"title": "cat2"}],
        "count": 2,
    }


@pytest.mark.asyncio
async def test_async_no_pagination_without_query_params():
    """
    Test that AsyncNoPagination works without any query parameters.
    Tests the async branch of the bug fix from https://github.com/vitalik/django-ninja/issues/1564

    AsyncNoPagination doesn't define any Input fields (uses empty PaginationBase.Input),
    so it should NOT require any query parameters.
    """
    api = NinjaAPI()

    @api.get("/items_no_pagination_async", response=List[int])
    @paginate(AsyncNoPagination)
    async def items_no_pagination_async(request):
        await asyncio.sleep(0)
        return ITEMS

    client = TestAsyncClient(api)

    response = await client.get("/items_no_pagination_async")
    assert response.status_code == 200
    result = response.json()
    assert result == {"count": 100, "items": ITEMS}

    # Check OpenAPI schema - should have no required parameters
    schema = api.get_openapi_schema()["paths"]["/api/items_no_pagination_async"]["get"]
    params = schema.get("parameters", [])

    # If there are any parameters, they should all be optional
    for param in params:
        assert (
            param.get("required", False) is False
        ), f"Parameter {param['name']} should not be required"



import datetime
from ninja import Schema, Path


class PathDate(Schema):
    year: int
    month: int
    day: int

    def value(self):
        return datetime.date(self.year, self.month, self.day)


@api.get("/events/{year}/{month}/{day}")
def events(request, date: Path[PathDate]):
    return {"date": date.value()}



import asyncio

import pytest

from ninja import NinjaAPI
from ninja.security import APIKeyQuery, HttpBearer
from ninja.testing import TestAsyncClient, TestClient


@pytest.mark.asyncio
async def test_async_view_handles_async_auth_func():
    api = NinjaAPI()

    async def auth(request):
        key = request.GET.get("key")
        if key == "secret":
            return key

    @api.get("/async", auth=auth)
    async def view(request):
        await asyncio.sleep(0)
        return {"key": request.auth}

    client = TestAsyncClient(api)

    # Actual tests --------------------------------------------------

    # without auth:
    res = await client.get("/async")
    assert res.status_code == 401

    # async successful
    res = await client.get("/async?key=secret")
    assert res.json() == {"key": "secret"}


@pytest.mark.asyncio
async def test_async_view_handles_async_auth_cls():
    api = NinjaAPI()

    class Auth:
        async def __call__(self, request):
            key = request.GET.get("key")
            if key == "secret":
                return key

    @api.get("/async", auth=Auth())
    async def view(request):
        await asyncio.sleep(0)
        return {"key": request.auth}

    client = TestAsyncClient(api)

    # Actual tests --------------------------------------------------

    # without auth:
    res = await client.get("/async")
    assert res.status_code == 401

    # async successful
    res = await client.get("/async?key=secret")
    assert res.json() == {"key": "secret"}


@pytest.mark.asyncio
async def test_async_view_handles_multi_auth():
    api = NinjaAPI()

    def auth_1(request):
        return None

    async def auth_2(request):
        return None

    async def auth_3(request):
        key = request.GET.get("key")
        if key == "secret":
            return key

    @api.get("/async", auth=[auth_1, auth_2, auth_3])
    async def view(request):
        await asyncio.sleep(0)
        return {"key": request.auth}

    client = TestAsyncClient(api)

    res = await client.get("/async?key=secret")
    assert res.json() == {"key": "secret"}


@pytest.mark.asyncio
async def test_async_view_handles_auth_errors():
    api = NinjaAPI()

    async def auth(request):
        raise Exception("boom")

    @api.get("/async", auth=auth)
    async def view(request):
        await asyncio.sleep(0)
        return {"key": request.auth}

    @api.exception_handler(Exception)
    def on_custom_error(request, exc):
        return api.create_response(request, {"custom": True}, status=401)

    client = TestAsyncClient(api)

    res = await client.get("/async?key=secret")
    assert res.json() == {"custom": True}


@pytest.mark.asyncio
async def test_sync_authenticate_method():
    class KeyAuth(APIKeyQuery):
        async def authenticate(self, request, key):
            await asyncio.sleep(0)
            if key == "secret":
                return key

    api = NinjaAPI(auth=KeyAuth())

    @api.get("/async")
    async def async_view(request):
        return {"auth": request.auth}

    client = TestAsyncClient(api)

    res = await client.get("/async")  # NO key
    assert res.json() == {"detail": "Unauthorized"}

    res = await client.get("/async?key=secret")
    assert res.json() == {"auth": "secret"}


def test_async_authenticate_method_in_sync_context():
    class KeyAuth(APIKeyQuery):
        async def authenticate(self, request, key):
            await asyncio.sleep(0)
            if key == "secret":
                return key

    api = NinjaAPI(auth=KeyAuth())

    @api.get("/sync")
    def sync_view(request):
        return {"auth": request.auth}

    client = TestClient(api)

    res = client.get("/sync")  # NO key
    assert res.json() == {"detail": "Unauthorized"}

    res = client.get("/sync?key=secret")
    assert res.json() == {"auth": "secret"}


@pytest.mark.asyncio
async def test_async_with_bearer():
    class BearerAuth(HttpBearer):
        async def authenticate(self, request, key):
            await asyncio.sleep(0)
            if key == "secret":
                return key

    api = NinjaAPI(auth=BearerAuth())

    @api.get("/async")
    async def async_view(request):
        return {"auth": request.auth}

    client = TestAsyncClient(api)

    res = await client.get("/async")  # NO key
    assert res.json() == {"detail": "Unauthorized"}

    res = await client.get("/async", headers={"Authorization": "Bearer secret"})
    assert res.json() == {"auth": "secret"}



from pathlib import Path

BASE_DIR = Path(__file__).parent.parent


SECRET_KEY = "NOT-SUPER-SECRET-DO-NOT-USE-ME"


DEBUG = True

ALLOWED_HOSTS = []


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "ninja",
    "someapp",
    "multi_param",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "demo.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "demo.wsgi.application"


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


STATIC_URL = "/static/"



from unittest.mock import patch

from ninja import NinjaAPI
from ninja.testing import TestClient


def test_examples():
    api = NinjaAPI()

    with patch("builtins.api", api, create=True):
        import docs.src.tutorial.path.code01  # noqa: F401

        client = TestClient(api)

        response = client.get("/items/123")
        assert response.json() == {"item_id": "123"}

    api = NinjaAPI()

    with patch("builtins.api", api, create=True):
        import docs.src.tutorial.path.code010  # noqa: F401
        import docs.src.tutorial.path.code02  # noqa: F401

        client = TestClient(api)

        response = client.get("/items/123")
        assert response.json() == {"item_id": 123}

        response = client.get("/events/2020/1/1")
        assert response.json() == {"date": "2020-01-01"}
        schema = api.get_openapi_schema(path_prefix="")
        events_params = schema["paths"]["/events/{year}/{month}/{day}"]["get"][
            "parameters"
        ]
        assert events_params == [
            {
                "in": "path",
                "name": "year",
                "schema": {"title": "Year", "type": "integer"},
                "required": True,
            },
            {
                "in": "path",
                "name": "month",
                "schema": {"title": "Month", "type": "integer"},
                "required": True,
            },
            {
                "in": "path",
                "name": "day",
                "schema": {"title": "Day", "type": "integer"},
                "required": True,
            },
        ]



from typing import List

import pytest
from django.core.files.uploadedfile import SimpleUploadedFile
from django.utils.datastructures import MultiValueDict

from ninja import File, NinjaAPI, UploadedFile
from ninja.errors import ConfigError
from ninja.testing import TestClient

api = NinjaAPI()


@api.post("/file1")
def file1(request, file: UploadedFile = File(...)):
    return {"name": file.name, "data": file.read().decode()}


@api.post("/file2")
def file_no_marker(request, file: UploadedFile):
    return {"name": file.name, "data": file.read().decode()}


@api.post("/file3")
def file_no_marker2(request, file: UploadedFile = None):
    return {"data": file and file.read().decode() or None}


@api.post("/file4")
def file_no_marker4(request, files: List[UploadedFile]):
    return {"result": [f.read().decode() for f in files]}


@api.post("/file5")
def file_no_marker5(request, file1: UploadedFile, file2: UploadedFile):
    return {"result": [f.read().decode() for f in (file1, file2)]}


@api.post("/file6")
def file_no_marker6(request, file: UploadedFile, files: List[UploadedFile]):
    return {"result": [f.read().decode() for f in [file] + files]}


client = TestClient(api)


def test_files():
    response = client.post("/file1")  # no file
    assert response.status_code == 422

    file = SimpleUploadedFile("test.txt", b"data123")
    response = client.post("/file1", FILES={"file": file})
    assert response.status_code == 200
    assert response.json() == {"name": "test.txt", "data": "data123"}

    file = SimpleUploadedFile("test.txt", b"data345")
    response = client.post("/file2", FILES={"file": file})
    assert response.status_code == 200, response.content
    assert response.json() == {"name": "test.txt", "data": "data345"}

    file = SimpleUploadedFile("test.txt", b"data567")
    response = client.post("/file3")
    assert response.status_code == 200, response.content
    assert response.json() == {"data": None}

    file = SimpleUploadedFile("test.txt", b"data789")
    response = client.post("/file4", FILES=MultiValueDict({"files": [file]}))
    assert response.status_code == 200, response.content
    assert response.json() == {"result": ["data789"]}

    file1 = SimpleUploadedFile("test1.txt", b"dataABC")
    file2 = SimpleUploadedFile("test2.txt", b"dataDEF")
    response = client.post("/file5", FILES={"file1": file1, "file2": file2})
    assert response.status_code == 200, response.content
    assert response.json() == {"result": ["dataABC", "dataDEF"]}

    file1 = SimpleUploadedFile("test1.txt", b"dataABC")
    file2 = SimpleUploadedFile("test2.txt", b"dataDEF")
    file3 = SimpleUploadedFile("test2.txt", b"dataGHI")
    response = client.post(
        "/file6", FILES=MultiValueDict({"file": [file1], "files": [file2, file3]})
    )
    assert response.status_code == 200, response.content
    assert response.json() == {"result": ["dataABC", "dataDEF", "dataGHI"]}


def test_schema():
    schema = api.get_openapi_schema()
    methods = []
    for pth in ["/file1", "/file2", "/file3", "/file4"]:
        method = schema["paths"][f"/api{pth}"]["post"]
        method = method["requestBody"]["content"]["multipart/form-data"]["schema"]
        methods.append(method)

    assert methods == [
        {
            "type": "object",
            "properties": {
                "file": {"type": "string", "format": "binary", "title": "File"}
            },
            "required": ["file"],
            "title": "FileParams",
        },
        {
            "type": "object",
            "properties": {
                "file": {"type": "string", "format": "binary", "title": "File"}
            },
            "required": ["file"],
            "title": "FileParams",
        },
        {
            "type": "object",
            "properties": {
                "file": {"type": "string", "format": "binary", "title": "File"}
            },
            "title": "FileParams",
        },
        {
            "type": "object",
            "properties": {
                "files": {
                    "type": "array",
                    "items": {"type": "string", "format": "binary"},
                    "title": "Files",
                }
            },
            "required": ["files"],
            "title": "FileParams",
        },
    ]


def test_invalid_file():
    with pytest.raises(ValueError):
        UploadedFile._validate("not_a_file", None)


def test_files_fix_middleware():
    api = NinjaAPI()

    with pytest.raises(ConfigError):

        @api.patch("/file1")
        def patch_with_file(request, file: UploadedFile):
            return {"name": file.name}



from ninja import NinjaAPI

api = NinjaAPI()


@api.get("/operation1", operation_id="my_id")
def operation_1(request):
    """
    This will be in description
    """
    return {"docstrings": True}


@api.get("/operation2", description="description from argument", deprecated=True)
def operation2(request):
    return {"description": True, "deprecated": True}


@api.get("/operation3", summary="Summary from argument", description="description arg")
def operation3(request):
    "This one also has docstring description"
    return {"summary": True, "description": "multiple"}


@api.get("/operation4", tags=["tag1", "tag2"])
def operation4(request):
    return {"tags": True}


@api.get(
    "/operation5",
    openapi_extra={
        "requestBody": {
            "content": {
                "application/json": {
                    "schema": {
                        "required": ["email"],
                        "type": "object",
                        "properties": {
                            "name": {"type": "string"},
                            "phone": {"type": "number"},
                            "email": {"type": "string"},
                        },
                    }
                }
            },
            "required": True,
        },
    },
)
def operation5(request):
    return {"openapi_extra": True}


@api.get("/not-included", include_in_schema=False)
def not_included(request):
    return True


schema = api.get_openapi_schema()


def test_schema():
    op1 = schema["paths"]["/api/operation1"]["get"]
    op2 = schema["paths"]["/api/operation2"]["get"]
    op3 = schema["paths"]["/api/operation3"]["get"]
    op4 = schema["paths"]["/api/operation4"]["get"]
    op5 = schema["paths"]["/api/operation5"]["get"]

    assert op1["operationId"] == "my_id"
    assert op2["operationId"] == "test_openapi_params_operation2"
    assert op3["operationId"] == "test_openapi_params_operation3"
    assert op4["operationId"] == "test_openapi_params_operation4"
    assert op5["operationId"] == "test_openapi_params_operation5"

    assert op1["summary"] == "Operation 1"
    assert op2["summary"] == "Operation2"
    assert op3["summary"] == "Summary from argument"
    assert op4["summary"] == "Operation4"
    assert op5["summary"] == "Operation5"

    assert op1["description"] == "This will be in description"

    assert op2["description"] == "description from argument"
    assert op2["deprecated"] is True

    assert op3["description"] == "description arg"

    assert op4["tags"] == ["tag1", "tag2"]

    assert op5["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "required": ["email"],
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "phone": {"type": "number"},
                        "email": {"type": "string"},
                    },
                }
            }
        },
        "required": True,
    }


def test_not_included():
    assert list(schema["paths"].keys()) == [
        "/api/operation1",
        "/api/operation2",
        "/api/operation3",
        "/api/operation4",
        "/api/operation5",
    ]
    # checking that "/not-included" is not there



import pytest

from ninja import NinjaAPI, Query
from ninja.utils import contribute_operation_args, replace_path_param_notation


@pytest.mark.parametrize(
    "input,expected_output",
    [
        ("abc/{def}", "abc/<def>"),
        ("abc/<def>", "abc/<def>"),
        ("abc", "abc"),
        ("<abc>", "<abc>"),
        ("{abc}", "<abc>"),
        ("{abc}/{def}", "<abc>/<def>"),
    ],
)
def test_replace_path_param_notation(input, expected_output):
    assert replace_path_param_notation(input) == expected_output


def test_contribute_operation_args():
    def some_func():
        pass

    contribute_operation_args(some_func, "arg1", str, Query(...))
    contribute_operation_args(some_func, "arg2", int, Query(...))

    api = NinjaAPI()

    api.get("/test")(some_func)

    schema = api.get_openapi_schema()
    assert schema["paths"]["/api/test"]["get"]["parameters"] == [
        {
            "in": "query",
            "name": "arg1",
            "schema": {"title": "Arg1", "type": "string"},
            "required": True,
        },
        {
            "in": "query",
            "name": "arg2",
            "schema": {"title": "Arg2", "type": "integer"},
            "required": True,
        },
    ]



from ninja import NinjaAPI, Form
from ninja.security import HttpBearer


class GlobalAuth(HttpBearer):
    def authenticate(self, request, token):
        if token == "supersecret":
            return token


api = NinjaAPI(auth=GlobalAuth())

# @api.get(...)
# def ...
# @api.post(...)
# def ...


@api.post("/token", auth=None)  # < overriding global auth
def get_token(request, username: str = Form(...), password: str = Form(...)):
    if username == "admin" and password == "giraffethinnknslong":
        return {"token": "supersecret"}



import re
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from django.urls import URLPattern
from django.urls import path as django_path
from django.utils.module_loading import import_string

from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.decorators import DecoratorMode
from ninja.errors import ConfigError
from ninja.operation import PathView
from ninja.throttling import BaseThrottle
from ninja.types import TCallable
from ninja.utils import normalize_path, replace_path_param_notation

if TYPE_CHECKING:
    from ninja import NinjaAPI  # pragma: no cover


__all__ = ["Router"]


class Router:
    def __init__(
        self,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        tags: Optional[List[str]] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
    ) -> None:
        self.api: Optional[NinjaAPI] = None
        self.auth = auth
        self.throttle = throttle
        self.tags = tags
        self.by_alias = by_alias
        self.exclude_unset = exclude_unset
        self.exclude_defaults = exclude_defaults
        self.exclude_none = exclude_none

        self.path_operations: Dict[str, PathView] = {}
        self._routers: List[Tuple[str, Router]] = []
        self._decorators: List[Tuple[Callable, DecoratorMode]] = []

    def get(
        self,
        path: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[TCallable], TCallable]:
        return self.api_operation(
            ["GET"],
            path,
            auth=auth,
            throttle=throttle,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )

    def post(
        self,
        path: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[TCallable], TCallable]:
        return self.api_operation(
            ["POST"],
            path,
            auth=auth,
            throttle=throttle,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )

    def delete(
        self,
        path: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[TCallable], TCallable]:
        return self.api_operation(
            ["DELETE"],
            path,
            auth=auth,
            throttle=throttle,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )

    def patch(
        self,
        path: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[TCallable], TCallable]:
        return self.api_operation(
            ["PATCH"],
            path,
            auth=auth,
            throttle=throttle,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )

    def put(
        self,
        path: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[TCallable], TCallable]:
        return self.api_operation(
            ["PUT"],
            path,
            auth=auth,
            throttle=throttle,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )

    def api_operation(
        self,
        methods: List[str],
        path: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[TCallable], TCallable]:
        def decorator(view_func: TCallable) -> TCallable:
            self.add_api_operation(
                path,
                methods,
                view_func,
                auth=auth,
                throttle=throttle,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                openapi_extra=openapi_extra,
            )
            return view_func

        return decorator

    def add_api_operation(
        self,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: Optional[bool] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> None:
        path = re.sub(r"\{uuid:(\w+)\}", r"{uuidstr:\1}", path, flags=re.IGNORECASE)
        # django by default convert strings to UUIDs
        # but we want to keep them as strings to let pydantic handle conversion/validation
        # if user whants UUID object
        # uuidstr is custom registered converter

        # No decoration here - will be done in build_routers

        if path not in self.path_operations:
            path_view = PathView()
            self.path_operations[path] = path_view
        else:
            path_view = self.path_operations[path]

        by_alias = by_alias is None and self.by_alias or by_alias
        exclude_unset = exclude_unset is None and self.exclude_unset or exclude_unset
        exclude_defaults = (
            exclude_defaults is None and self.exclude_defaults or exclude_defaults
        )
        exclude_none = exclude_none is None and self.exclude_none or exclude_none

        path_view.add_operation(
            path=path,
            methods=methods,
            view_func=view_func,
            auth=auth,
            throttle=throttle,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )
        if self.api:
            path_view.set_api_instance(self.api, self)

        return None

    def set_api_instance(
        self, api: "NinjaAPI", parent_router: Optional["Router"] = None
    ) -> None:
        if self.auth is NOT_SET and parent_router:
            self.auth = parent_router.auth
        self.api = api

        for path_view in self.path_operations.values():
            path_view.set_api_instance(self.api, self)
        for _, router in self._routers:
            router.set_api_instance(api, self)

    def urls_paths(self, prefix: str) -> Iterator[URLPattern]:
        # Ensure decorators are applied before generating URLs
        self._apply_decorators_to_operations()

        prefix = replace_path_param_notation(prefix)
        for path, path_view in self.path_operations.items():
            for operation in path_view.operations:
                path = replace_path_param_notation(path)
                route = "/".join([i for i in (prefix, path) if i])
                # to skip lot of checks we simply treat double slash as a mistake:
                route = normalize_path(route)
                route = route.lstrip("/")

                url_name = getattr(operation, "url_name", "")
                if not url_name and self.api:
                    url_name = self.api.get_operation_url_name(operation, router=self)

                yield django_path(route, path_view.get_view(), name=url_name)

    def add_router(
        self,
        prefix: str,
        router: Union["Router", str],
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        tags: Optional[List[str]] = None,
    ) -> None:
        if isinstance(router, str):
            router = import_string(router)
            assert isinstance(router, Router)

        if self.api:
            # we are already attached to an api
            self.api.add_router(
                prefix=prefix,
                router=router,
                auth=auth,
                throttle=throttle,
                tags=tags,
                parent_router=self,
            )
        else:
            # we are not attached to an api
            if auth != NOT_SET:
                router.auth = auth
            # TODO: throttle
            if tags is not None:
                router.tags = tags
            self._routers.append((prefix, router))

    def add_decorator(
        self,
        decorator: Callable,
        mode: DecoratorMode = "operation",
    ) -> None:
        """
        Add a decorator to be applied to all operations in this router.

        Args:
            decorator: The decorator function to apply
            mode: "operation" (default) applies after validation,
                  "view" applies before validation
        """
        if mode not in ("view", "operation"):
            raise ValueError(f"Invalid decorator mode: {mode}")
        self._decorators.append((decorator, mode))

    def build_routers(self, prefix: str) -> List[Tuple[str, "Router"]]:
        if self.api is not None:
            from ninja.main import debug_server_url_reimport

            if not debug_server_url_reimport():
                raise ConfigError(
                    f"Router@'{prefix}' has already been attached to API"
                    f" {self.api.title}:{self.api.version} "
                )

        # Apply decorators to all operations in this router
        self._apply_decorators_to_operations()

        internal_routes = []
        for inter_prefix, inter_router in self._routers:
            # Inherit decorators from parent router
            # Prepend parent decorators so they execute first (outer decorators)
            inter_router._decorators = self._decorators + inter_router._decorators
            _route = normalize_path("/".join((prefix, inter_prefix))).lstrip("/")
            internal_routes.extend(inter_router.build_routers(_route))

        return [(prefix, self), *internal_routes]

    def _apply_decorators_to_operations(self) -> None:
        """Apply all stored decorators to operations in this router"""
        for path_view in self.path_operations.values():
            for operation in path_view.operations:
                # Track what decorators have already been applied to avoid duplicates
                applied_decorators = getattr(operation, "_applied_decorators", [])

                # Apply decorators that haven't been applied yet
                for decorator, mode in self._decorators:
                    if (decorator, mode) not in applied_decorators:
                        if mode == "view":
                            operation.run = decorator(operation.run)  # type: ignore
                        elif mode == "operation":
                            operation.view_func = decorator(operation.view_func)
                        else:
                            raise ValueError(
                                f"Invalid decorator mode: {mode}"
                            )  # pragma: no cover
                        applied_decorators.append((decorator, mode))

                # Store what decorators have been applied
                operation._applied_decorators = applied_decorators  # type: ignore[attr-defined]



import json
from enum import Enum
from ipaddress import IPv4Address, IPv6Address
from typing import List, Union

import pytest
from django.http import HttpResponse
from pydantic import BaseModel, ValidationError
from pydantic_core import Url

from ninja import Router
from ninja.responses import Response
from ninja.testing import TestClient

router = Router()


@router.get("/check_int", response=int)
def check_int(request):
    return "1"


@router.get("/check_int2", response=int)
def check_int2(request):
    return "str"


class User:
    def __init__(self, id, user_name, password):
        self.id = id
        self.user_name = user_name
        self.password = password


class MyEnum(Enum):
    first = "first"
    second = "second"


def to_camel(string: str) -> str:
    words = string.split("_")
    return words[0].lower() + "".join(word.capitalize() for word in words[1:])


class UserModel(BaseModel):
    id: int
    user_name: str
    # skipping password output to responses

    model_config = dict(
        from_attributes=True,
        alias_generator=to_camel,
        populate_by_name=True,
    )


@router.get("/check_model", response=UserModel)
def check_model(request):
    return User(1, "John", "Password")


@router.get("/check_list_model", response=List[UserModel])
def check_list_model(request):
    return [User(1, "John", "Password")]


@router.get("/check_model_alias", response=UserModel, by_alias=True)
def check_model_alias(request):
    return User(1, "John", "Password")


@router.get("/check_union", response=Union[int, UserModel])
def check_union(request, q: int):
    if q == 0:
        return 1
    if q == 1:
        return User(1, "John", "Password")
    return "invalid"


@router.get("/check_set_header")
def check_set_header(request, response: HttpResponse):
    response["Cache-Control"] = "no-cache"
    return 1


@router.get("/check_set_cookie")
def check_set_cookie(request, set: bool, response: HttpResponse):
    if set:
        response.set_cookie("test", "me")
    return 1


@router.get("/check_del_cookie")
def check_del_cookie(request, response: HttpResponse):
    response.delete_cookie("test")
    return 1


client = TestClient(router)


@pytest.mark.parametrize(
    "path,expected_response",
    [
        ("/check_int", 1),
        ("/check_model", {"id": 1, "user_name": "John"}),  # the password is skipped
        (
            "/check_list_model",
            [{"id": 1, "user_name": "John"}],
        ),  # the password is skipped
        ("/check_model", {"id": 1, "user_name": "John"}),  # the password is skipped
        ("/check_model_alias", {"id": 1, "userName": "John"}),  # result is camelCase
        ("/check_union?q=0", 1),
        ("/check_union?q=1", {"id": 1, "user_name": "John"}),
    ],
)
def test_responses(path, expected_response):
    response = client.get(path)
    assert response.status_code == 200, response.content
    assert response.json() == expected_response
    assert response.data == response.data == expected_response  # Ensures cache works


def test_validates():
    with pytest.raises(ValidationError):
        client.get("/check_int2")

    with pytest.raises(ValidationError):
        client.get("/check_union?q=2")


def test_set_header():
    response = client.get("/check_set_header")
    assert response.status_code == 200
    assert response.content == b"1"
    assert response["Cache-Control"] == "no-cache"


def test_set_cookie():
    response = client.get("/check_set_cookie?set=0")
    assert "test" not in response.cookies

    response = client.get("/check_set_cookie?set=1")
    cookie = response.cookies.get("test")
    assert cookie
    assert cookie.value == "me"


def test_del_cookie():
    response = client.get("/check_del_cookie")
    cookie = response.cookies.get("test")
    assert cookie
    assert cookie["expires"] == "Thu, 01 Jan 1970 00:00:00 GMT"
    assert cookie["max-age"] == 0


def test_ipv4address_encoding():
    data = {"ipv4": IPv4Address("127.0.0.1")}
    response = Response(data)
    response_data = json.loads(response.content)
    assert response_data["ipv4"] == str(data["ipv4"])


def test_ipv6address_encoding():
    data = {"ipv6": IPv6Address("::1")}
    response = Response(data)
    response_data = json.loads(response.content)
    assert response_data["ipv6"] == str(data["ipv6"])


def test_enum_encoding():
    data = {"enum": MyEnum.first}
    response = Response(data)
    response_data = json.loads(response.content)
    assert response_data["enum"] == str(data["enum"])


def test_pydantic_url():
    data = {"url": Url("https://django-ninja.dev/")}
    response = Response(data)
    response_data = json.loads(response.content)
    assert response_data == {"url": "https://django-ninja.dev/"}



from functools import partial
from typing import TYPE_CHECKING, Any, List

from django.urls import path

from .views import default_home, openapi_json, openapi_view

if TYPE_CHECKING:
    from ninja import NinjaAPI  # pragma: no cover

__all__ = ["get_openapi_urls", "get_root_url"]


def get_openapi_urls(api: "NinjaAPI") -> List[Any]:
    result = []

    if api.openapi_url:
        view = partial(openapi_json, api=api)
        if api.docs_decorator:
            view = api.docs_decorator(view)  # type: ignore
        result.append(
            path(api.openapi_url.lstrip("/"), view, name="openapi-json"),
        )

        assert (
            api.openapi_url != api.docs_url
        ), "Please use different urls for openapi_url and docs_url"

        if api.docs_url:
            view = partial(openapi_view, api=api)
            if api.docs_decorator:
                view = api.docs_decorator(view)  # type: ignore
            result.append(
                path(api.docs_url.lstrip("/"), view, name="openapi-view"),
            )

    return result


def get_root_url(api: "NinjaAPI") -> Any:
    return path("", partial(default_home, api=api), name="api-root")



import itertools
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple, Type, Union, cast

from django.db.models import Field as DjangoField
from django.db.models import ManyToManyRel, ManyToOneRel, Model
from pydantic import create_model as create_pydantic_model

from ninja.errors import ConfigError
from ninja.orm.fields import get_schema_field
from ninja.schema import Schema

# MAYBE:
# Schema = create_schema(Model, exclude=['id'])
#
# @api.post
# def operation_create(request, payload: Schema):
#     orm_instance = payload.orm.apply(Model())
#     orm_instance.save()
#
# @api.post("/{id}")
# def operation_edit(request, id: int, payload: Schema):
#     orm_instance = payload.orm.apply(Model.objects.get(id=id))
#     orm_instance.save()

__all__ = ["SchemaFactory", "factory", "create_schema"]

SchemaKey = Tuple[Type[Model], str, int, str, str, str, str]


class SchemaFactory:
    def __init__(self) -> None:
        self.schemas: Dict[SchemaKey, Type[Schema]] = {}
        self.schema_names: Set[str] = set()

    def create_schema(
        self,
        model: Type[Model],
        *,
        name: str = "",
        depth: int = 0,
        fields: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        optional_fields: Optional[List[str]] = None,
        custom_fields: Optional[List[Tuple[str, Any, Any]]] = None,
        base_class: Type[Schema] = Schema,
    ) -> Type[Schema]:
        name = name or model.__name__

        if fields and exclude:
            raise ConfigError("Only one of 'fields' or 'exclude' should be set.")

        key = self.get_key(
            model, name, depth, fields, exclude, optional_fields, custom_fields
        )
        if key in self.schemas:
            return self.schemas[key]

        model_fields_list = list(self._selected_model_fields(model, fields, exclude))
        if optional_fields:
            if optional_fields == "__all__":
                optional_fields = [f.name for f in model_fields_list]

        definitions = {}
        for fld in model_fields_list:
            python_type, field_info = get_schema_field(
                fld,
                depth=depth,
                optional=optional_fields and (fld.name in optional_fields),
            )
            definitions[fld.name] = (python_type, field_info)

        if custom_fields:
            for fld_name, python_type, field_info in custom_fields:
                # if not isinstance(field_info, FieldInfo):
                #     field_info = Field(field_info)
                definitions[fld_name] = (python_type, field_info)

        if name in self.schema_names:
            name = self._get_unique_name(name)

        schema: Type[Schema] = create_pydantic_model(
            name,
            __config__=None,
            __base__=base_class,
            __module__=base_class.__module__,
            __validators__={},
            **definitions,
        )  # type: ignore
        # __model_name: str,
        # *,
        # __config__: ConfigDict | None = None,
        # __base__: None = None,
        # __module__: str = __name__,
        # __validators__: dict[str, AnyClassMethod] | None = None,
        # __cls_kwargs__: dict[str, Any] | None = None,
        # **field_definitions: Any,
        self.schemas[key] = schema
        self.schema_names.add(name)
        return schema

    def get_key(
        self,
        model: Type[Model],
        name: str,
        depth: int,
        fields: Union[str, List[str], None],
        exclude: Optional[List[str]],
        optional_fields: Optional[Union[List[str], str]],
        custom_fields: Optional[List[Tuple[str, str, Any]]],
    ) -> SchemaKey:
        "returns a hashable value for all given parameters"
        # TODO: must be a test that compares all kwargs from init to get_key
        return (
            model,
            name,
            depth,
            str(fields),
            str(exclude),
            str(optional_fields),
            str(custom_fields),
        )

    def _get_unique_name(self, name: str) -> str:
        "Returns a unique name by adding counter suffix"
        for num in itertools.count(start=2):  # pragma: no branch
            result = f"{name}{num}"
            if result not in self.schema_names:
                break
        return result

    def _selected_model_fields(
        self,
        model: Type[Model],
        fields: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ) -> Iterator[DjangoField]:
        "Returns iterator for model fields based on `exclude` or `fields` arguments"
        all_fields = {f.name: f for f in self._model_fields(model)}

        if not fields and not exclude:
            for f in all_fields.values():
                yield f

        invalid_fields = (set(fields or []) | set(exclude or [])) - all_fields.keys()
        if invalid_fields:
            raise ConfigError(
                f"DjangoField(s) {invalid_fields} are not in model {model}"
            )

        if fields:
            for name in fields:
                yield all_fields[name]
        if exclude:
            for f in all_fields.values():
                if f.name not in exclude:
                    yield f

    def _model_fields(self, model: Type[Model]) -> Iterator[DjangoField]:
        "returns iterator with all the fields that can be part of schema"
        for fld in model._meta.get_fields():
            if isinstance(fld, (ManyToOneRel, ManyToManyRel)):
                # skipping relations
                continue
            yield cast(DjangoField, fld)


factory = SchemaFactory()

create_schema = factory.create_schema



from typing import List, Optional

import pytest

from ninja import NinjaAPI, Schema
from ninja.patch_dict import PatchDict
from ninja.testing import TestClient

api = NinjaAPI()

client = TestClient(api)


class SomeSchema(Schema):
    name: str
    age: int
    category: Optional[str] = None


class OtherSchema(SomeSchema):
    other: str
    category: Optional[List[str]] = None


@api.patch("/patch")
def patch(request, payload: PatchDict[SomeSchema]):
    return {"payload": payload, "type": str(type(payload))}


@api.patch("/patch-inherited")
def patch_inherited(request, payload: PatchDict[OtherSchema]):
    return {"payload": payload, "type": str(type(payload))}


@pytest.mark.parametrize(
    "input,output",
    [
        ({"name": "foo"}, {"name": "foo"}),
        ({"age": "1"}, {"age": 1}),
        ({}, {}),
        ({"wrong_param": 1}, {}),
        ({"age": None}, {"age": None}),
    ],
)
def test_patch_calls(input: dict, output: dict):
    response = client.patch("/patch", json=input)
    assert response.json() == {"payload": output, "type": "<class 'dict'>"}


def test_schema():
    "Checking that json schema properties are all optional"
    schema = api.get_openapi_schema()
    assert schema["components"]["schemas"]["SomeSchemaPatch"] == {
        "title": "SomeSchemaPatch",
        "type": "object",
        "properties": {
            "name": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "title": "Name",
            },
            "age": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "title": "Age",
            },
            "category": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "title": "Category",
            },
        },
    }


def test_patch_inherited():
    input = {"other": "any", "category": ["cat1", "cat2"]}
    expected_output = {"payload": input, "type": "<class 'dict'>"}

    response = client.patch("/patch-inherited", json=input)
    assert response.json() == expected_output


def test_inherited_schema():
    "Checking that json schema properties for inherithed schemas are ok"
    schema = api.get_openapi_schema()
    assert schema["components"]["schemas"]["OtherSchemaPatch"] == {
        "title": "OtherSchemaPatch",
        "type": "object",
        "properties": {
            "name": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "title": "Name",
            },
            "age": {
                "anyOf": [{"type": "integer"}, {"type": "null"}],
                "title": "Age",
            },
            "other": {
                "anyOf": [{"type": "string"}, {"type": "null"}],
                "title": "Other",
            },
            "category": {
                "anyOf": [
                    {
                        "items": {
                            "type": "string",
                        },
                        "type": "array",
                    },
                    {"type": "null"},
                ],
                "title": "Category",
            },
        },
    }



from typing import Optional

from ninja.params.models import DictStrAny, ParamModel


class _NestedParamModel(ParamModel):
    outer: DictStrAny
    leaf: Optional[int]

    __ninja_flatten_map__ = {
        "foo": ("outer", "foo"),
        "bar": ("outer", "bar"),
        "leaf": ("leaf",),
    }


def test_map_data_paths_creates_parent_for_missing_nested_values():
    assert _NestedParamModel._map_data_paths({}) == {"outer": {}}


def test_map_data_paths_sets_values_when_present():
    data = _NestedParamModel._map_data_paths({"foo": 1, "leaf": 2})
    assert data == {"outer": {"foo": 1}, "leaf": 2}



import copy
import uuid

import pytest
from pydantic import BaseModel

from ninja import NinjaAPI
from ninja.constants import NOT_SET
from ninja.signature.details import is_pydantic_model
from ninja.signature.utils import UUIDStrConverter
from ninja.testing import TestClient


def test_is_pydantic_model():
    class Model(BaseModel):
        x: int

    assert is_pydantic_model(Model)
    assert is_pydantic_model("instance") is False


def test_client():
    "covering everything in testclient (including invalid paths)"
    api = NinjaAPI()
    client = TestClient(api)
    with pytest.raises(Exception):  # noqa: B017
        client.get("/404")


def test_kwargs():
    api = NinjaAPI()

    @api.get("/")
    def operation(request, a: str, *args, **kwargs):
        pass

    schema = api.get_openapi_schema()
    params = schema["paths"]["/api/"]["get"]["parameters"]
    print(params)
    assert params == [  # Only `a` should be here, not kwargs
        {
            "in": "query",
            "name": "a",
            "schema": {"title": "A", "type": "string"},
            "required": True,
        }
    ]


def test_uuid_converter():
    conv = UUIDStrConverter()
    assert isinstance(conv.to_url(uuid.uuid4()), str)


def test_copy_not_set():
    assert id(NOT_SET) == id(copy.copy(NOT_SET))
    assert id(NOT_SET) == id(copy.deepcopy(NOT_SET))



from ninja.security import HttpBearer


class AuthBearer(HttpBearer):
    def authenticate(self, request, token):
        if token == "supersecret":
            return token


@api.get("/bearer", auth=AuthBearer())
def bearer(request):
    return {"token": request.auth}



import re

from django.conf import settings
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt, ensure_csrf_cookie

from ninja import NinjaAPI
from ninja.security import APIKeyCookie, APIKeyHeader, django_auth
from ninja.testing import TestClient as BaseTestClient


class AnyCookieAuth(APIKeyCookie):
    """A mock authentication class that accepts any cookie value.
    To test CSRF functionality without specific authentication logic.
    """

    def authenticate(self, request, key):
        return True


class TestClient(BaseTestClient):
    """
    A mock authentication class that accepts any cookie value.
    To test CSRF functionality without specific authentication logic.
    """

    def _build_request(self, *args, **kwargs):
        request = super()._build_request(*args, **kwargs)
        request._dont_enforce_csrf_checks = False
        return request


csrf_OFF = NinjaAPI(urls_namespace="csrf_OFF")
csrf_ON = NinjaAPI(urls_namespace="csrf_ON", auth=AnyCookieAuth())  # , csrf=True
csrf_ON_with_django_auth = NinjaAPI(urls_namespace="csrf_ON", auth=django_auth)


@csrf_OFF.post("/post")
def post_off(request):
    return {"success": True}


@csrf_ON.post("/post")
def post_on(request):
    return {"success": True}


@csrf_ON.post("/post/csrf_exempt")
@csrf_exempt
def post_on_with_exempt(request):
    return {"success": True}


TOKEN = "1bcdefghij2bcdefghij3bcdefghij4bcdefghij5bcdefghij6bcdefghijABCD"
COOKIES = {settings.CSRF_COOKIE_NAME: TOKEN}


def test_csrf_off():
    client = TestClient(csrf_OFF)
    assert client.post("/post", COOKIES=COOKIES).status_code == 200


def test_csrf_on():
    client = TestClient(csrf_ON)

    assert client.post("/post", COOKIES=COOKIES).status_code == 403

    # check with token in formdata
    response = client.post("/post", {"csrfmiddlewaretoken": TOKEN}, COOKIES=COOKIES)
    assert response.status_code == 200

    # check with headers
    response = client.post("/post", COOKIES=COOKIES, headers={"X-CSRFTOKEN": TOKEN})
    assert response.status_code == 200

    # exempt check
    assert client.post("/post/csrf_exempt", COOKIES=COOKIES).status_code == 200


def test_csrf_cookie_auth():
    "Cookie based authtentication should have csrf check by default"

    class CookieAuth(APIKeyCookie):
        def authenticate(self, request, key):
            return key == "test"

    cookie_auth = CookieAuth()
    api = NinjaAPI(auth=cookie_auth)

    @api.post("/test")
    def test_view(request):
        return {"success": True}

    client = TestClient(api)

    # No auth - access denied
    assert client.post("/test").status_code == 403

    # Cookie auth + valid csrf
    cookies = {"key": "test"}
    cookies.update(COOKIES)
    response = client.post("/test", COOKIES=cookies, headers={"X-CSRFTOKEN": TOKEN})
    assert response.status_code == 200, response.content

    # Cookie auth + INVALID csrf
    response = client.post(
        "/test", COOKIES=cookies, headers={"X-CSRFTOKEN": TOKEN + "invalid"}
    )
    assert response.status_code == 403, response.content

    # Turning off csrf on cookie, valid key, no csrf passed
    cookie_auth.csrf = False
    response = client.post("/test", COOKIES={"key": "test"})
    assert response.status_code == 200, response.content


def test_csrf_cookies_can_be_obtained():
    @csrf_ON.get("/obtain_csrf_token_get")
    @ensure_csrf_cookie
    def obtain_csrf_token_get(request):
        return JsonResponse(data={"success": True})

    @csrf_ON.post("/obtain_csrf_token_post")
    @ensure_csrf_cookie
    @csrf_exempt
    def obtain_csrf_token_post(request):
        return JsonResponse(data={"success": True})

    @csrf_ON_with_django_auth.get("/obtain_csrf_token_get", auth=None)
    @ensure_csrf_cookie
    def obtain_csrf_token_get_no_auth_route(request):
        return JsonResponse(data={"success": True})

    @csrf_ON_with_django_auth.post("/obtain_csrf_token_post", auth=None)
    @ensure_csrf_cookie
    @csrf_exempt
    def obtain_csrf_token_post_no_auth_route(request):
        return JsonResponse(data={"success": True})

    client = TestClient(csrf_ON)
    # can get csrf cookie through get
    response = client.get("/obtain_csrf_token_get")
    assert response.status_code == 200
    assert len(response.cookies["csrftoken"].value) > 0
    # can get csrf cookie through exempted post
    response = client.post("/obtain_csrf_token_post")
    assert response.status_code == 200
    assert len(response.cookies["csrftoken"].value) > 0
    # Now testing a route with disabled auth from a client with django_auth set globally also works
    client = TestClient(csrf_ON_with_django_auth)
    # can get csrf cookie through get on route with disabled auth
    response = client.get("/obtain_csrf_token_get")
    assert response.status_code == 200
    assert len(response.cookies["csrftoken"].value) > 0
    # can get csrf cookie through exempted post on route with disabled auth
    response = client.post("/obtain_csrf_token_post")
    assert response.status_code == 200
    assert len(response.cookies["csrftoken"].value) > 0


def test_docs():
    "Testing that docs are initializing csrf headers correctly"

    api = NinjaAPI(auth=AnyCookieAuth())

    client = TestClient(api)
    resp = client.get("/docs")
    assert resp.status_code == 200
    csrf_token = re.findall(r'data-csrf-token="(.*?)"', resp.content.decode("utf8"))[0]
    assert len(csrf_token) > 0


def test_no_auth_csrf_exempt():
    """Test that APIs without authentication are CSRF exempt by default"""
    from django.middleware.csrf import CsrfViewMiddleware
    from django.test import RequestFactory

    api = NinjaAPI(urls_namespace="test_no_auth_csrf")

    @api.post("/create")
    def create_item(request):
        return {"status": "created"}

    # Get the actual view function that Django will use
    patterns = api.urls[0]
    view_func = None
    for pattern in patterns:
        if hasattr(pattern, "callback") and "create" in str(pattern.pattern):
            view_func = pattern.callback
            break

    assert view_func is not None, "Could not find view function"

    # Test 1: Check if view has csrf_exempt attribute
    has_csrf_exempt = hasattr(view_func, "csrf_exempt")
    csrf_exempt_value = getattr(view_func, "csrf_exempt", None)
    print(f"View has csrf_exempt: {has_csrf_exempt}, value: {csrf_exempt_value}")

    # Test 2: Simulate Django's CSRF middleware check
    factory = RequestFactory()
    request = factory.post("/create", data="{}", content_type="application/json")

    csrf_middleware = CsrfViewMiddleware(lambda r: None)
    csrf_middleware.process_request(request)
    csrf_response = csrf_middleware.process_view(request, view_func, (), {})

    # If csrf_response is None, the request passed CSRF checks
    # If it's not None, it's a 403 Forbidden response
    assert (
        csrf_response is None
    ), f"CSRF middleware blocked the request! Regular APIs should be CSRF exempt. Response: {csrf_response}"


def test_docs_cookie_auth():
    class CookieAuth(APIKeyCookie):
        def authenticate(self, request, key):
            return key == "test"

    class HeaderAuth(APIKeyHeader):
        def authenticate(self, request, key):
            return key == "test"

    api = NinjaAPI(auth=CookieAuth())
    client = TestClient(api)
    resp = client.get("/docs")
    csrf_token = re.findall(r'data-csrf-token="(.*?)"', resp.content.decode("utf8"))[0]
    assert len(csrf_token) > 0

    api = NinjaAPI(auth=HeaderAuth())
    client = TestClient(api)
    resp = client.get("/docs")
    csrf_token = re.findall(r'data-csrf-token="(.*?)"', resp.content.decode("utf8"))[0]
    assert len(csrf_token) == 0



from ninja import Schema


class Item(Schema):
    name: str
    description: str = None
    price: float
    quantity: int


@api.put("/items/{item_id}")
def update(request, item_id: int, item: Item):
    return {"item_id": item_id, "item": item.dict()}



from typing import List

import pytest

from ninja import NinjaAPI, Schema
from ninja.pagination import PageNumberPagination, RouterPaginated, paginate
from ninja.testing import TestAsyncClient, TestClient

api = NinjaAPI(default_router=RouterPaginated())


class ItemSchema(Schema):
    id: int


@api.get("/items", response=List[ItemSchema])
def items(request):
    return [{"id": i} for i in range(1, 51)]


@api.get("/items_nolist", response=ItemSchema)
def items_nolist(request):
    return {"id": 1}


@api.get("/items_extra_layer", response=List[ItemSchema])
@paginate  # has not to break down
def items_extra_layer(request):
    return [{"id": i} for i in range(1, 51)]


@api.get("/items_overridden", response=List[ItemSchema])
@paginate(PageNumberPagination, page_size=3)  # has precedence over router pagination
def items_overridden_pagination(request):
    return [{"id": i} for i in range(1, 51)]


client = TestClient(api)


def test_for_list_reponse():
    parameters = api.get_openapi_schema()["paths"]["/api/items"]["get"]["parameters"]
    assert parameters == [
        {
            "in": "query",
            "name": "limit",
            "schema": {
                "title": "Limit",
                "default": 100,
                "minimum": 1,
                "type": "integer",
            },
            "required": False,
        },
        {
            "in": "query",
            "name": "offset",
            "schema": {
                "title": "Offset",
                "default": 0,
                "minimum": 0,
                "type": "integer",
            },
            "required": False,
        },
    ]

    response = client.get("/items?offset=5&limit=1").json()
    # print(response)
    assert response == {"items": [{"id": 6}], "count": 50}


def test_for_NON_list_reponse():
    parameters = api.get_openapi_schema()["paths"]["/api/items_nolist"]["get"][
        "parameters"
    ]
    # print(parameters)
    assert parameters == []


def test_extra_pagination_layer_does_not_crash():
    parameters = api.get_openapi_schema()["paths"]["/api/items_extra_layer"]["get"][
        "parameters"
    ]
    assert parameters == [
        {
            "in": "query",
            "name": "limit",
            "schema": {
                "title": "Limit",
                "default": 100,
                "minimum": 1,
                "type": "integer",
            },
            "required": False,
        },
        {
            "in": "query",
            "name": "offset",
            "schema": {
                "title": "Offset",
                "default": 0,
                "minimum": 0,
                "type": "integer",
            },
            "required": False,
        },
    ]

    response = client.get("/items_extra_layer?offset=5&limit=1").json()
    assert response == {"items": [{"id": 6}], "count": 50}


def test_for_list_with_overridden_pagination_reponse():
    parameters = api.get_openapi_schema()["paths"]["/api/items_overridden"]["get"][
        "parameters"
    ]
    assert parameters == [
        {
            "in": "query",
            "name": "page",
            "schema": {
                "title": "Page",
                "default": 1,
                "minimum": 1,
                "type": "integer",
            },
            "required": False,
        },
        {
            "in": "query",
            "name": "page_size",
            "schema": {
                "anyOf": [{"minimum": 1, "type": "integer"}, {"type": "null"}],
                "title": "Page Size",
            },
            "required": False,
        },
    ]

    response = client.get("/items_overridden?page=5").json()
    assert response == {"items": [{"id": 13}, {"id": 14}, {"id": 15}], "count": 50}


@pytest.mark.asyncio
async def test_async_pagination():
    @api.get("/items_async", response=List[ItemSchema])
    async def items_async(request):
        return [{"id": i} for i in range(1, 51)]

    client = TestAsyncClient(api)

    response = await client.get("/items_async?offset=5&limit=1")
    assert response.json() == {"items": [{"id": 6}], "count": 50}



import asyncio
from functools import wraps

import pytest

from ninja import NinjaAPI, Router
from ninja.testing import TestAsyncClient, TestClient


# Async test decorators
def async_operation_decorator(func):
    """Async decorator that adds data after validation (operation level)"""

    @wraps(func)
    async def wrapper(request, *args, **kwargs):
        result = await func(request, *args, **kwargs)
        if isinstance(result, dict):
            result["async_operation_decorated"] = True
        return result

    return wrapper


def async_view_decorator(func):
    """Async decorator that adds a header before validation (view level)"""

    @wraps(func)
    async def wrapper(request, *args, **kwargs):
        response = await func(request, *args, **kwargs)
        response["X-Async-View-Decorator"] = "applied"
        return response

    return wrapper


@pytest.mark.asyncio
async def test_router_add_decorator_async_operation_mode():
    """Test add_decorator on router with async operations in OPERATION mode"""
    api = NinjaAPI()
    router = Router()

    # Add decorator to router
    router.add_decorator(async_operation_decorator, mode="operation")

    @router.get("/test")
    async def endpoint(request):
        await asyncio.sleep(0)  # Simulate async work
        return {"message": "async test"}

    api.add_router("/", router)
    client = TestAsyncClient(api)

    response = await client.get("/test")
    assert response.status_code == 200
    assert response.json() == {
        "message": "async test",
        "async_operation_decorated": True,
    }


@pytest.mark.asyncio
async def test_router_add_decorator_async_view_mode():
    """Test add_decorator on router with async operations in VIEW mode"""
    api = NinjaAPI()
    router = Router()

    # Add decorator to router
    router.add_decorator(async_view_decorator, mode="view")

    @router.get("/test")
    async def endpoint(request):
        await asyncio.sleep(0)
        return {"message": "async test"}

    api.add_router("/", router)
    client = TestAsyncClient(api)

    response = await client.get("/test")
    assert response.status_code == 200
    assert response["X-Async-View-Decorator"] == "applied"
    assert response.json() == {"message": "async test"}


@pytest.mark.asyncio
async def test_api_add_decorator_async():
    """Test add_decorator on API with async operations"""
    api = NinjaAPI()

    # Add decorator to entire API
    api.add_decorator(async_operation_decorator, mode="operation")

    @api.get("/test1")
    async def endpoint1(request):
        await asyncio.sleep(0)
        return {"message": "test1"}

    @api.get("/test2")
    async def endpoint2(request):
        await asyncio.sleep(0)
        return {"message": "test2"}

    client = TestAsyncClient(api)

    # Both endpoints should be decorated
    response = await client.get("/test1")
    assert response.status_code == 200
    assert response.json() == {"message": "test1", "async_operation_decorated": True}

    response = await client.get("/test2")
    assert response.status_code == 200
    assert response.json() == {"message": "test2", "async_operation_decorated": True}


@pytest.mark.asyncio
async def test_mixed_sync_async_decorators():
    """Test that sync decorators work with async endpoints"""
    api = NinjaAPI()
    router = Router()

    # Use a sync decorator on async endpoint
    def sync_decorator(func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            # For async functions, this will return a coroutine
            result = func(request, *args, **kwargs)
            if asyncio.iscoroutine(result):

                async def async_wrapper():
                    actual_result = await result
                    if isinstance(actual_result, dict):
                        actual_result["sync_decorated"] = True
                    return actual_result

                return async_wrapper()
            else:
                # For sync functions, modify the result directly
                if isinstance(result, dict):
                    result["sync_decorated"] = True
                return result

        return wrapper

    router.add_decorator(sync_decorator, mode="operation")

    @router.get("/async")
    async def async_endpoint(request):
        await asyncio.sleep(0)
        return {"type": "async"}

    @router.get("/sync")
    def sync_endpoint(request):
        return {"type": "sync"}

    api.add_router("/", router)
    client = TestAsyncClient(api)

    # Test async endpoint
    response = await client.get("/async")
    assert response.status_code == 200
    assert response.json() == {"type": "async", "sync_decorated": True}

    # Test sync endpoint with regular TestClient
    from ninja.testing import TestClient

    sync_client = TestClient(api)
    response = sync_client.get("/sync")
    assert response.status_code == 200
    assert response.json() == {"type": "sync", "sync_decorated": True}


@pytest.mark.asyncio
async def test_mixed_sync_async_endpoints_same_router():
    """Test router with both sync and async endpoints using the same decorator"""
    api = NinjaAPI()
    router = Router()

    # Universal decorator that works with both sync and async functions
    def universal_decorator(func):
        if asyncio.iscoroutinefunction(func):
            # Handle async functions
            @wraps(func)
            async def async_wrapper(request, *args, **kwargs):
                result = await func(request, *args, **kwargs)
                if isinstance(result, dict):
                    result["universal_decorated"] = True
                    result["func_type"] = "async"
                return result

            return async_wrapper
        else:
            # Handle sync functions
            @wraps(func)
            def sync_wrapper(request, *args, **kwargs):
                result = func(request, *args, **kwargs)
                if isinstance(result, dict):
                    result["universal_decorated"] = True
                    result["func_type"] = "sync"
                return result

            return sync_wrapper

    router.add_decorator(universal_decorator, mode="operation")

    @router.get("/async")
    async def async_endpoint(request):
        await asyncio.sleep(0)
        return {"endpoint": "async"}

    @router.get("/sync")
    def sync_endpoint(request):
        return {"endpoint": "sync"}

    api.add_router("/", router)

    # Test both endpoints with appropriate clients
    async_client = TestAsyncClient(api)
    sync_client = TestClient(api)

    # Test async endpoint
    response = await async_client.get("/async")
    assert response.status_code == 200
    assert response.json() == {
        "endpoint": "async",
        "universal_decorated": True,
        "func_type": "async",
    }

    # Test sync endpoint
    response = sync_client.get("/sync")
    assert response.status_code == 200
    assert response.json() == {
        "endpoint": "sync",
        "universal_decorated": True,
        "func_type": "sync",
    }



import hashlib
import time
from typing import Dict, List, Optional, Tuple

from django.core.cache import cache as default_cache
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest


class BaseThrottle:
    """
    Rate throttling of requests.
    """

    def allow_request(self, request: HttpRequest) -> bool:
        """
        Return `True` if the request should be allowed, `False` otherwise.
        """
        raise NotImplementedError(".allow_request() must be overridden")

    def get_ident(self, request: HttpRequest) -> Optional[str]:
        """
        Identify the machine making the request by parsing HTTP_X_FORWARDED_FOR
        if present and number of proxies is > 0. If not use all of
        HTTP_X_FORWARDED_FOR if it is available, if not use REMOTE_ADDR.
        """
        from ninja.conf import settings

        xff = request.META.get("HTTP_X_FORWARDED_FOR")
        remote_addr = request.META.get("REMOTE_ADDR")
        num_proxies = settings.NUM_PROXIES

        if num_proxies is not None:
            if num_proxies == 0 or xff is None:
                return remote_addr
            addrs: List[str] = xff.split(",")
            client_addr = addrs[-min(num_proxies, len(addrs))]
            return client_addr.strip()

        return "".join(xff.split()) if xff else remote_addr

    def wait(self) -> Optional[float]:
        """
        Optionally, return a recommended number of seconds to wait before
        the next request.
        """
        return None


class SimpleRateThrottle(BaseThrottle):
    """
    A simple cache implementation, that only requires `.get_cache_key()`
    to be overridden.

    The rate (requests / seconds) is set by a `rate` attribute on the Throttle
    class.  The attribute is a string of the form 'number_of_requests/period'.

    Period should be one of: ('s', 'sec', 'm', 'min', 'h', 'hour', 'd', 'day')

    Previous request information used for throttling is stored in the cache.
    """

    from ninja.conf import settings

    cache = default_cache
    timer = time.time
    cache_format = "throttle_%(scope)s_%(ident)s"
    scope: Optional[str] = None
    THROTTLE_RATES: Dict[str, Optional[str]] = settings.DEFAULT_THROTTLE_RATES
    _PERIODS = {
        "s": 1,
        "m": 60,
        "h": 60 * 60,
        "d": 60 * 60 * 24,
        "sec": 1,
        "min": 60,
        "hour": 60 * 60,
        "day": 60 * 60 * 24,
    }

    def __init__(self, rate: Optional[str] = None):
        self.rate: Optional[str]
        if rate:
            self.rate = rate
        else:
            self.rate = self.get_rate()
        self.num_requests, self.duration = self.parse_rate(self.rate)

    def get_cache_key(self, request: HttpRequest) -> Optional[str]:
        """
        Should return a unique cache-key which can be used for throttling.
        Must be overridden.

        May return `None` if the request should not be throttled.
        """
        raise NotImplementedError(".get_cache_key() must be overridden")

    def get_rate(self) -> Optional[str]:
        """
        Determine the string representation of the allowed request rate.
        """
        if not getattr(self, "scope", None):
            msg = f"You must set either `.scope` or `.rate` for '{self.__class__.__name__}' throttle"
            raise ImproperlyConfigured(msg)

        try:
            return self.THROTTLE_RATES[self.scope]  # type: ignore
        except KeyError:
            msg = f"No default throttle rate set for '{self.scope}' scope"
            raise ImproperlyConfigured(msg) from None

    def parse_rate(self, rate: Optional[str]) -> Tuple[Optional[int], Optional[int]]:
        """
        Given the request rate string, return a two tuple of:
        <allowed number of requests>, <period of time in seconds>
        """
        if rate is None:
            return (None, None)

        try:
            count, rest = rate.split("/", 1)

            for unit in self._PERIODS:
                if rest.endswith(unit):
                    multi = int(rest[: -len(unit)]) if rest[: -len(unit)] else 1
                    period = self._PERIODS[unit]
                    break
            else:
                multi, period = int(rest), 1

            return int(count), multi * period

        except (ValueError, IndexError):
            raise ValueError(f"Invalid rate format: {rate}") from None

    def allow_request(self, request: HttpRequest) -> bool:
        """
        Implement the check to see if the request should be throttled.

        On success calls `throttle_success`.
        On failure calls `throttle_failure`.
        """
        # if self.rate is None:
        #     return True

        self.key = self.get_cache_key(request)
        if self.key is None:
            return True

        self.history = self.cache.get(self.key, [])
        self.now = self.timer()  # type: ignore

        # Drop any requests from the history which have now passed the
        # throttle duration
        while self.history and self.history[-1] <= self.now - self.duration:  # type: ignore
            self.history.pop()
        if len(self.history) >= self.num_requests:  # type: ignore
            return self.throttle_failure()
        return self.throttle_success()

    def throttle_success(self) -> bool:
        """
        Inserts the current request's timestamp along with the key
        into the cache.
        """
        self.history.insert(0, self.now)
        self.cache.set(self.key, self.history, self.duration)
        return True

    def throttle_failure(self) -> bool:
        """
        Called when a request to the API has failed due to throttling.
        """
        return False

    def wait(self) -> Optional[float]:
        """
        Returns the recommended next request time in seconds.
        """
        if self.history:
            remaining_duration = self.duration - (self.now - self.history[-1])
        else:
            remaining_duration = self.duration

        available_requests = self.num_requests - len(self.history) + 1  # type: ignore
        if available_requests <= 0:
            return None

        return remaining_duration / float(available_requests)  # type: ignore


class AnonRateThrottle(SimpleRateThrottle):
    """
    Limits the rate of API calls that may be made by a anonymous users.

    The IP address of the request will be used as the unique cache key.
    """

    scope = "anon"

    def get_cache_key(self, request: HttpRequest) -> Optional[str]:
        if getattr(request, "auth", None) is not None:
            return None  # Only throttle unauthenticated requests.

        return self.cache_format % {
            "scope": self.scope,
            "ident": self.get_ident(request),
        }


class AuthRateThrottle(SimpleRateThrottle):
    """
    Limits the rate of API calls that may be made by a given user.

    The string representation of request.auth object will be used as a unique cache key.
    If you use custom auth objects make sure to implement __str__ method.
    For anonymous requests, the IP address of the request will be used.
    """

    scope = "auth"

    def get_cache_key(self, request: HttpRequest) -> str:
        if getattr(request, "auth", None) is not None:
            ident = hashlib.sha256(str(request.auth).encode()).hexdigest()  # type: ignore
            # TODO: ^maybe auth should have an attribute that developer can overwrite
        else:
            ident = self.get_ident(request)  # type: ignore

        return self.cache_format % {"scope": self.scope, "ident": ident}


class UserRateThrottle(SimpleRateThrottle):
    """
    Limits the rate of API calls that may be made by a given user.

    The user id will be used as a unique cache key if the user is
    authenticated.  For anonymous requests, the IP address of the request will
    be used.
    """

    scope = "user"

    def get_cache_key(self, request: HttpRequest) -> str:
        if request.user and request.user.is_authenticated:
            ident = request.user.pk
        else:
            ident = self.get_ident(request)

        return self.cache_format % {"scope": self.scope, "ident": ident}
